#ifndef CERECONNHELPERS_H_324EA174
#define CERECONNHELPERS_H_324EA174
#include <boost/integer.hpp>

#ifdef CC_TESTING_DEBUG
#define FRIEND_TEST(test_case_name, test_name)\
friend class test_case_name##_##test_name##_Test
#endif

namespace cc
{
    template <typename T, int NumBits>
    using boost_integer_type = 
        typename std::conditional<
            std::is_unsigned<T>::value,
            boost::uint_t<NumBits>, 
            boost::int_t<NumBits>
        >::type;

    template <typename T>
    constexpr unsigned size_in_bits() { return sizeof(T) * CHAR_BIT; }

    template <typename T>
    using smallest_larger_integral_t = 
        typename boost_integer_type<T, size_in_bits<T>() + 1>::least;


    template<typename InputIt1, typename InputIt2>
    bool it_chain_equal(InputIt1 first1, InputIt1 last1, InputIt2 first2, InputIt2 last2)
    {
        // loop through each element of old and new channel list
        for (; first1 != last1 and first2 != last2; ++ first1, ++ first2)
        {
            // if there's a mismatch, lists are different
            if (*first1 != *first2)
            {
                return false;
            }
            // if so far no mismatch and both list entries are 0, lists are equal
            if (*first1 == 0 && *first2 == 0)
            {
                return true;
            }
        }
        return (((first1 == last1) || (*first1 == 0)) && ((first2 == last2) || (*first2 == 0) ));
    }
    
    template<typename T, typename U>
    bool it_chain_equal(const T list1, const U list2)
    {
        return it_chain_equal(list1.cbegin(), list1.cend(), list2.cbegin(), list2.cend());
    }
    
    
} /* cc */

#endif /* end of include guard: CERECONNHELPERS_H_324EA174 */
