#ifndef CERECONSTANTS_H_0C754D41
#define CERECONSTANTS_H_0C754D41

/** \file CereConstants.h
    \brief  CereConn common constants
    
    This file defines constants in use by the CereConn module.
    Some constants are derived from cbhwlib.h (CereLink).
    In those cases, the original constant is given as comment
    at the end of the line.
    
© Copyright 2018  - Jonas B. Zimmermann. All Rights Reserved.

    \author Jonas B. Zimmermann 
    \date 2018-08-23

**/
#include <array>

namespace cc
{
    //! Namespace encapsulating constants for use with cbSDK
    namespace constants
    {
        constexpr unsigned int MaxFEChannels        = 128;      //!< cbNUM_FE_CHANS
        constexpr unsigned int MaxAnalogChannels    = 144;      //!< cbNUM_ANALOG_CHANS
        constexpr unsigned int MaxUnits             = 6;
        constexpr unsigned int ContSamples          = 60000;
        constexpr unsigned int NSampleGroups        = 6;        //!< cbMAXGROUPS, but only using 6 as seems conventional
        constexpr std::array<unsigned int, NSampleGroups> SampleGroups          = {{500, 1000, 2000, 10000, 30000, 30000}};
        constexpr std::array<unsigned int, NSampleGroups> SampleGroupPeriods    = {{60, 30, 15, 3, 1, 1}};
        
        constexpr unsigned int SpikesCached         = 30000;
        constexpr unsigned int CommentsCached       = 1000;
        constexpr unsigned int DigInEvCached        = 1000;
        constexpr unsigned int EventSnippetSamples  = 128;      //!< cbMAX_PNTS
        constexpr unsigned int CommentLength        = 128;      //!< cbMAX_COMMENT
        
        constexpr unsigned int ChannelDigitalIn     = 151;      //!< MAX_CHANS_DIGITAL_IN
        constexpr unsigned int ChannelSerialIn      = 152;      //!< MAX_CHANS_SERIAL
        
        
    }
}

#endif /* end of include guard: CERECONSTANTS_H_0C754D41 */
