#ifndef COMMENTRINGBUFFER_H_C2C59ECF
#define COMMENTRINGBUFFER_H_C2C59ECF

#include "cbhwlib.h"

// Structures and classes to store comments
namespace cc
{

template<typename TTimestamp, size_t CommentLength>
struct CommentT
{
    TTimestamp    timestamp;
    uint8_t     flag;
    uint32_t    data;
    char        comment[CommentLength];
    
    CommentT()
    : timestamp(0)
    , flag(0)
    , data(0)
    , comment("")
    {
    }
    
    CommentT(const cbPKT_COMMENT & pkt_comment)
        : timestamp(pkt_comment.time)
        , flag(pkt_comment.info.flags)
        , data(pkt_comment.data)
    {
        static_assert(sizeof(comment) == sizeof(pkt_comment.comment), "comment fields must match in length");
        std::copy(std::begin(pkt_comment.comment), std::end(pkt_comment.comment), std::begin(comment));
    }
    
};


template<typename TTimestamp, size_t CommentLength, size_t Size>
class CommentRingBufferT: public RingBufferT<CommentT<TTimestamp, CommentLength>, Size>
{
public:
    bool addComment(const cbPKT_COMMENT &pkt_comment);
};

template<typename TTimestamp, size_t CommentLength, size_t Size>
bool CommentRingBufferT<TTimestamp, CommentLength, Size>::addComment(const cbPKT_COMMENT & pkt_comment)
{
    if (!this->wasFull())
    {
        CommentT<TTimestamp, CommentLength> comment(pkt_comment);
        return this->push(comment);
    }
    ++ (this->m_overflow_count);
    return false;  // full queue
}

template<typename TTimestamp, size_t CommentLength>
using CommentListT = std::vector<CommentT<TTimestamp, CommentLength>>;

}
#endif /* end of include guard: COMMENTRINGBUFFER_H_C2C59ECF */
