#ifndef DIGINEVRINGBUFFER_H_07183E4B
#define DIGINEVRINGBUFFER_H_07183E4B
#include <cstdio>

#include "cbhwlib.h"

// Structures and classes to store digital input events
namespace cc
{

template<typename TTimestamp>
struct DigInEvT
{
    enum class Source
    {
        Unknown = -1,
        DigIn   = 0,
        Serial  = 1
    };
    
    TTimestamp  timestamp;
    Source      source;
    uint32_t    data;
    
    DigInEvT()
    : timestamp(0)
    , source(Source::Unknown)
    , data(0)
    {
    }
    
    DigInEvT(const cbPKT_DINP & pkt_digin)
        : timestamp(pkt_digin.time)
        , source(Source::Unknown)
        , data(pkt_digin.data[0])
    {
        if (pkt_digin.chid == constants::ChannelSerialIn)
        {
            source = Source::Serial;
        }
        else if (pkt_digin.chid == constants::ChannelDigitalIn)
        {
            source = Source::DigIn;
        }
    }
};


template<typename TTimestamp, size_t Size>
class DigInEvRingBufferT: public RingBufferT<DigInEvT<TTimestamp>, Size>
{
public:
    bool addDigIn(const cbPKT_DINP &pkt_digin);
};

template<typename TTimestamp, size_t Size>
bool DigInEvRingBufferT<TTimestamp, Size>::addDigIn(const cbPKT_DINP & pkt_digin)
{
    if (!this->wasFull())
    {
        DigInEvT<TTimestamp> digIn(pkt_digin);
        return this->push(digIn);
    }
    ++ (this->m_overflow_count);
    return false;  // full queue
}

template<typename TTimestamp>
using DigInEvListT = std::vector<DigInEvT<TTimestamp>>;

}
#endif /* end of include guard: DIGINEVRINGBUFFER_H_07183E4B */
