#ifndef SPIKERINGBUFFER_H_FB0D8F9E
#define SPIKERINGBUFFER_H_FB0D8F9E

#include <array>

namespace cc
{
template<typename TData, typename TTimestamp, typename TChannel, typename TUnit, size_t SnippetLength>
struct SpikeT
{
    using SnippetT = std::array<TData, SnippetLength>;
    
    TTimestamp      timestamp;
    TChannel        channel;
    TUnit           unit;
    SnippetT        snippet;
    
    SpikeT()
    : timestamp(0)
    , unit(0)
    {
    }
    
    SpikeT(TTimestamp ts, TChannel chid, TUnit u)
    : timestamp(ts)
    , channel(chid)
    , unit(u)
    {
    }
    
    SpikeT(const cbPKT_SPK & pkt_spk)
    : timestamp(pkt_spk.time)
    , channel(pkt_spk.chid)
    , unit(pkt_spk.unit)
    {
        static_assert(SnippetLength == sizeof(pkt_spk.wave) / sizeof(pkt_spk.wave[0]), "spike snippet fields must match in length");
        std::copy(std::begin(pkt_spk.wave), std::end(pkt_spk.wave), std::begin(snippet));
    }
    
};

template<typename TData, typename TTimestamp, typename TChannel, typename TUnit, size_t SnippetLength, size_t Size>
class SpikeRingBufferT: public RingBufferT<SpikeT<TData, TTimestamp, TChannel, TUnit, SnippetLength>, Size>
{
public:
    bool addSpike(const cbPKT_SPK &pkt_spk);
};

template<typename TData, typename TTimestamp, typename TChannel, typename TUnit, size_t SnippetLength, size_t Size>
bool SpikeRingBufferT<TData, TTimestamp, TChannel, TUnit, SnippetLength, Size>::addSpike(const cbPKT_SPK & pkt_spk)
{
    if (!this->wasFull())
    {
        SpikeT<TData, TTimestamp, TChannel, TUnit, SnippetLength> spike(pkt_spk);
        return this->push(spike);
    }
    ++ (this->m_overflow_count);
    return false;  // full queue
}

}

#endif /* end of include guard: SPIKERINGBUFFER_H_FB0D8F9E */
