#ifndef SPIKERATEESTIMATIONMETHODS_IMPL_TPP_1E146BC4
#define SPIKERATEESTIMATIONMETHODS_IMPL_TPP_1E146BC4

#include <cmath>

namespace cc
{
namespace sre
{

template<class FwdIt, typename T>
T r_estimate_id(const FwdIt& begin_it, const FwdIt& end_it)
{
    return (T) *begin_it;
}

template<class Container, typename T>
T r_estimate_id_w_cont(const Container& c)
{
    return r_estimate_id(c.crbegin(), c.crend());
}


template<class FwdIt, typename T>
T r_estimate_with_exp_decay(const FwdIt& begin_it, const FwdIt& end_it, const T& decay_factor, const size_t max_length)
{
    auto it{begin_it};

    T val{0.0}, norm{0.0}, df{1.0};
    size_t count = 0;
    while ((it != end_it) && (count < max_length))
    {
        norm += df;
        val += *it * df;
        df *= decay_factor;
        ++ count;
        ++ it;
    }
    if (count == 0)
    {
        return T(0);
    }
    return val / norm;
}

template<class Container, typename T>
T r_estimate_with_exp_decay_w_cont(const Container& c, const T& decay_factor, const size_t max_length)
{
    return r_estimate_with_exp_decay(c.crbegin(), c.crend(), decay_factor, max_length);
}

template<class FwdIt, typename T>
T r_estimate_with_boxcar(const FwdIt& begin_it, const FwdIt& end_it, const size_t length)
{
    auto it{begin_it};
    
    T val{0};
    size_t count = 0;
    while (it != end_it && count < length)
    {
        val += *it;
        ++ count;
        ++ it;
    }
    if (count == 0)
    {
        return T(0);
    }
    return val / (T) count;
}

template<typename Container, typename T>
T r_estimate_with_boxcar_w_cont(const Container& c, const size_t length)
{
    return r_estimate_with_boxcar(c.crbegin(), c.crend(), length);
}

    
} // namespace sre
} // namespace cc
#endif /* end of include guard: SPIKERATEESTIMATIONMETHODS_IMPL_TPP_1E146BC4 */
