#ifndef SPIKERATEESTIMATOR_IMPL_TPP_C088933B
#define SPIKERATEESTIMATOR_IMPL_TPP_C088933B


namespace cc
{
namespace sre
{


template<typename T>
bool SpikeRateEstimator::addSpike(T && spk)
{
    if (validate_spk(spk))
    {
        // mp_logger->debug("Adding a spike for (ch: {}, unit: {})", spk.channel, spk.unit);
        return mp_sq->write(spk);
    }
    return false;
}

template<typename T>
bool SpikeRateEstimator::addSpike(const T& spk)
{
    if (validate_spk(spk))
    {
        // mp_logger->debug("Adding a spike for (ch: {}, unit: {})", spk.channel, spk.unit);
        return mp_sq->write(spk);
    }
    return false;
}

template<typename T>
inline bool SpikeRateEstimator::validate_spk(const T& spk)
{
    auto search = m_ch_un_map.find(std::make_pair(spk.channel, spk.unit));
    return search != m_ch_un_map.end();
}

bool SpikeRateEstimator::ch_u_offset(size_type ch, size_type u, size_type & unit_idx)
{
    auto search = m_ch_un_map.find(std::make_pair(ch,u));
    if (search != m_ch_un_map.end())
    {
        mp_logger->debug("Found (ch, u: {}, {}), at pos {}", search->first.first, search->first.second, search->second);
        unit_idx = search->second;
        return true;
    }
    else
    {
        mp_logger->debug("Not found (ch, u: {}, {})", ch, u);
        return false;
    }
}

SpikeRateEstimator::size_type SpikeRateEstimator::n_ch_u()
{
    return m_ch_un_list.size();
}


} // namespace sre
} // namespace cc

#endif /* end of include guard: SPIKERATEESTIMATOR_IMPL_TPP_C088933B */
