#ifndef UNITHIST_IMPL_TPP_DD319F33
#define UNITHIST_IMPL_TPP_DD319F33


namespace cc
{
namespace sre
{

// UnitHist::const_iterator implementation

// constructors and destructors
template <typename T, class Container>
UnitHist<T,Container>::const_iterator::const_iterator(const const_iterator& other)
: m_cont(other.m_cont)
, m_pos(other.m_pos)
, m_iterated(other.m_iterated)
{
}

template <typename T, class Container>
UnitHist<T,Container>::const_iterator::~const_iterator()
{
}

// private constructors
template <typename T, class Container>
UnitHist<T,Container>::const_iterator::const_iterator(const UnitHist& cont, container_const_iterator start)
: m_cont(cont)
, m_pos(start)
, m_iterated(false)
{
}

template <typename T, class Container>
UnitHist<T,Container>::const_iterator::const_iterator(const UnitHist& cont, container_const_iterator start, bool iterated)
: m_cont(cont)
, m_pos(start)
, m_iterated(iterated)
{
}

// public function members
template <typename T, class Container>
typename UnitHist<T,Container>::const_iterator& UnitHist<T,Container>::const_iterator::operator=(const const_iterator& other)
{
    if (!compatible(other))
    {
        throw std::invalid_argument("Iterators must refer to same object for assignment.");
    }
    m_pos = other.m_pos;
    m_iterated = other.m_iterated;
    return *this;
};

template <typename T, class Container>
bool UnitHist<T,Container>::const_iterator::operator==(const const_iterator& rhs) const
{
    return (m_pos == rhs.m_pos) && (m_iterated == rhs.m_iterated);
}

template <typename T, class Container>
bool UnitHist<T,Container>::const_iterator::operator!=(const const_iterator& rhs) const
{
    return !(*this == rhs);
}

template <typename T, class Container>
typename UnitHist<T,Container>::const_iterator& UnitHist<T,Container>::const_iterator::operator++()
{
    if (m_cont.size() == 0)
    {
        m_iterated = true;
        return *this;
    }

    do
    {
        if (m_pos == m_cont.m_c.cend() - 1)
        {
            m_pos = m_cont.m_c.cbegin();
        }
        else
        {
            ++ m_pos;
        }
    } while (m_pos == m_cont.m_c_it);
    
    m_iterated = true;
    return *this;
}

template <typename T, class Container>
typename UnitHist<T,Container>::const_iterator UnitHist<T,Container>::const_iterator::operator++(int)
{
    const_iterator tmp = *this;
    ++ *this;
    return tmp;
}

template <typename T, class Container>
typename UnitHist<T,Container>::const_iterator& UnitHist<T,Container>::const_iterator::operator--()
{
    if (m_cont.size() == 0)
    {
        m_iterated = true;
        return *this;
    }
    do
    {
        if (m_pos == m_cont.m_c.cbegin())
        {
            m_pos =  m_cont.m_c.cend() - 1;
        }
        else
        {
            -- m_pos;
        }
    } while (m_pos == m_cont.m_c_it);
    
    m_iterated = true;
    return *this;
}

template <typename T, class Container>
typename UnitHist<T,Container>::const_iterator UnitHist<T,Container>::const_iterator::operator--(int)
{
    const_iterator tmp = *this;
    -- *this;
    return tmp;
}

template <typename T, class Container>
typename UnitHist<T,Container>::const_iterator& UnitHist<T,Container>::const_iterator::operator+=(const_iterator::difference_type add)
{
    std::advance(*this, add);
    return *this;
}

template <typename T, class Container>
typename UnitHist<T,Container>::const_iterator UnitHist<T,Container>::const_iterator::operator+(const_iterator::difference_type add) const
{
    auto new_it = *this;
    new_it += add;
    return new_it;
}

template <typename T, class Container>
typename UnitHist<T,Container>::const_iterator& UnitHist<T,Container>::const_iterator::operator-=(const_iterator::difference_type sub)
{
    *this += -sub;
    return *this;
}

template <typename T, class Container>
typename UnitHist<T,Container>::const_iterator UnitHist<T,Container>::const_iterator::operator-(const_iterator::difference_type sub) const
{
    return *this + (-sub);
}

template <typename T, class Container>
typename UnitHist<T,Container>::const_iterator::difference_type UnitHist<T,Container>::const_iterator::operator-(const_iterator rhs) const
{
    if (!compatible(rhs))
    {
        throw std::invalid_argument("Iterators must refer to same object for difference.");
    }
    if (*this == rhs)
    {
        return 0;
    }
    
    if (rhs.m_pos != this->m_pos)
    {
        return  (this->m_pos - rhs.m_pos);
    }
    if (rhs.m_pos == this->m_pos)
    {
        if (rhs.m_iterated)
        {
            return -(difference_type)m_cont.size();
        }
        else
        {
            return (difference_type)m_cont.size();
        }
    }
    return 42;
}

template <typename T, class Container>
typename UnitHist<T,Container>::const_iterator::reference UnitHist<T,Container>::const_iterator::operator*() const
{
    return *m_pos;
}

template <typename T, class Container>
bool UnitHist<T,Container>::const_iterator::compatible(const const_iterator& other) const
{
    return (&m_cont == &(other.m_cont));
}


#ifdef CC_TESTING_DEBUG
template <typename T, class Container>
void UnitHist<T,Container>::const_iterator::print_debug()
{
    printf("It: m_c_it, m_pos, size, m_iterated: %li, %li, %li, %i\n", m_cont.m_c_it - m_cont.m_c.cbegin(), m_pos - m_cont.m_c.cbegin(), m_cont.m_c.size(), m_iterated);
}
#endif


// UnitHist implementation

// constructors
template <typename T, class Container>
UnitHist<T,Container>::UnitHist(const size_type c_size)
: m_size(c_size)
, m_container_size(m_size + 1)
, m_c(m_container_size, T())
, m_c_it(m_c.begin())
, m_zero_spikes(m_size)
{
}

template <typename T, class Container>
UnitHist<T, Container>::UnitHist(const Container& c)
: m_size(c.size())
, m_container_size(m_size + 1)
, m_c(m_container_size, T())
, m_c_it(m_c.begin())
{
    std::copy(c.cbegin(), c.cend(), m_c.begin() + 1);
    m_zero_spikes = std::find_if( m_c.crbegin(), m_c.crend() - 1, [](const value_type& v){return v != 0;}) - m_c.crbegin(); 
}

template <typename T, class Container>
UnitHist<T, Container>::UnitHist(Container&& c)
: m_size(c.size())
, m_container_size(m_size + 1)
, m_c(m_container_size, T())
, m_c_it(m_c.begin())
{
    std::copy(std::make_move_iterator(c.begin()), std::make_move_iterator(c.end()), m_c.begin() + 1);
    m_zero_spikes = std::find_if( m_c.crbegin(), m_c.crend() - 1, [](const value_type& v){return v != 0;}) - m_c.crbegin();
}

template <typename T, class Container>
UnitHist<T, Container>::UnitHist(const UnitHist<T, Container>& other)
: m_size(other.m_size)
, m_container_size(other.m_container_size)
, m_c(other.m_c)
, m_c_it(m_c.begin() + (other.m_c_it - other.m_c.cbegin()))
, m_zero_spikes(other.m_zero_spikes)
{
}

template <typename T, class Container>
UnitHist<T, Container>::UnitHist(UnitHist<T, Container>&& other)
: m_size(other.m_size)
, m_container_size(other.m_container_size)
, m_c(std::move(other.m_c))
, m_c_it(std::move(other.m_c_it))
, m_zero_spikes(other.m_zero_spikes)
{
}

template <typename T, class Container>
bool UnitHist<T, Container>::operator==(const UnitHist<T, Container> &other) const
{
    return std::equal(cbegin(), cend(), other.cbegin(), other.cend());
}

template <typename T, class Container>
bool UnitHist<T, Container>::operator!=(const UnitHist<T, Container> &other) const
{
    return !(*this == other);
}

template <typename T, class Container>
UnitHist<T, Container>& UnitHist<T, Container>::operator++()
{
    ++(*m_c_it);
    return *this;
}


template <typename T, class Container>
typename UnitHist<T, Container>::size_type UnitHist<T, Container>::size() const
{
    return m_size;
}

template <typename T, class Container>
void UnitHist<T, Container>::push(const UnitHist<T, Container>::value_type& x)
{
    *m_c_it = x;
    step();
}

template <typename T, class Container>
void UnitHist<T, Container>::push(UnitHist<T, Container>::value_type&& x)
{
    *m_c_it = x;
    step();
}

template <typename T, class Container>
UnitHist<T, Container>& UnitHist<T, Container>::operator<<(const UnitHist<T, Container>::value_type & x)
{
    push(x);
    return *this;
}

template <typename T, class Container>
UnitHist<T, Container>& UnitHist<T, Container>::operator<<(UnitHist<T, Container>::value_type&& x)
{
    push(x);
    return *this;
}

template <typename T, class Container>
void UnitHist<T, Container>::step()
{
    _step();
    *m_c_it = 0;
}

template <typename T, class Container>
void UnitHist<T, Container>::add_to_current(const UnitHist<T, Container>::value_type& x)
{
    if (x > 0)
    {
        (*m_c_it) += x;
    }
}

template <typename T, class Container>
void UnitHist<T, Container>::add_to_current(UnitHist<T, Container>::value_type&& x)
{
    if (x > 0)
    {
        (*m_c_it) += x;
    }
}

template <typename T, class Container>
bool UnitHist<T, Container>::has_spikes() const
{
    return m_zero_spikes < size();
}

template <typename T, class Container>
typename UnitHist<T, Container>::const_iterator UnitHist<T, Container>::begin() const
{
    if (m_c_it == m_c.cend() - 1)
    {
        return const_iterator(*this, m_c.cbegin(), size() == 0);
    }
    return const_iterator(*this, m_c_it + 1, false);
}

template <typename T, class Container>
typename UnitHist<T, Container>::const_iterator UnitHist<T, Container>::cbegin() const
{
    return begin();
}


template <typename T, class Container>
typename UnitHist<T, Container>::const_iterator UnitHist<T, Container>::end() const
{
    if (m_c_it == m_c.cend() - 1)
    {
        return const_iterator(*this, m_c.cbegin(), true);
    }
    return const_iterator(*this, m_c_it + 1, true);
}

template <typename T, class Container>
typename UnitHist<T, Container>::const_iterator UnitHist<T, Container>::cend() const
{
    return end();
}

template <typename T, class Container>
typename UnitHist<T, Container>::const_reverse_iterator UnitHist<T, Container>::rbegin() const
{
    if (m_c_it == m_c.cend() - 1)
    {
        return const_reverse_iterator(const_iterator(*this, m_c.begin(), size() == 0));
    }
    return const_reverse_iterator(const_iterator(*this, m_c_it + 1, false));
}

template <typename T, class Container>
typename UnitHist<T, Container>::const_reverse_iterator UnitHist<T, Container>::crbegin() const
{
    return rbegin();
}

template <typename T, class Container>
typename UnitHist<T, Container>::const_reverse_iterator UnitHist<T, Container>::rend() const
{
    if (m_c_it == m_c.cend() - 1)
    {
        return const_reverse_iterator(const_iterator(*this, m_c.cbegin(), true));
    }
    return const_reverse_iterator(const_iterator(*this, m_c_it + 1, true));
}

template <typename T, class Container>
typename UnitHist<T, Container>::const_reverse_iterator UnitHist<T, Container>::crend() const
{
    return rend();
}

template <typename T, class Container>
void UnitHist<T, Container>::_step()
{
    // if more than zero spikes, reset zero spike bin counter
    if (*m_c_it > 0)
    {
        m_zero_spikes = 0;
    }
    else
    {
        ++ m_zero_spikes;
    }
    
    ++ m_c_it;
    if (m_c_it == m_c.end())
    {
        m_c_it = m_c.begin();
    }
}

template <typename T,
          class Container>
std::ostream& operator<<(std::ostream& os, const UnitHist<T, Container>& c)
{ 
    os << "UnitHist: ";
    for (auto cit = c.crbegin(); cit != c.crend(); ++cit)
    {
        os << *cit << " ";
    }
    return os; 
}



} // namespace sre
} // namespace cc

#endif /* end of include guard: UNITHIST_IMPL_TPP_DD319F33 */
