# distutils: language=c++

from libc.stdint cimport uint8_t, int16_t, uint32_t
from libcpp.string cimport string
from libcpp.map cimport map
from libcpp.pair cimport pair
from libcpp.set cimport set
from libcpp.unordered_map cimport unordered_map
from libcpp.vector cimport vector
from libcpp cimport bool

cimport cere_constants as ccon
cimport arma_cy as ac

#cdef extern from "CereRingBuffer.h" namespace "cc::DigInEv":
cpdef enum Source:
    dies_Unknown = -1,
    dies_DigIn = 0,
    dies_Serial = 1
cdef extern from "CereConnBaseTypes.h" namespace "cc":
    ctypedef vector[uint32_t]    ChannelNumList_t
    
cdef extern from "CereRingBuffer.h" namespace "cc":
    cdef cppclass SampleGroupChunk:
        SampleGroupChunk() except +
        uint32_t                last_timestamp
        size_t                  sample_period
        vector[int16_t]         buffer
        vector[uint32_t]        channel_ids
        size_t                  overflow_count

    cdef cppclass SampleGroupList:
        SampleGroupList() except +
        uint32_t                    timestamp
        vector[SampleGroupChunk]    sgc

    cdef cppclass Comment:
        uint32_t    timestamp
        uint8_t     flag
        uint32_t    data
        char        comment[ccon.CommentLength+0]
        Comment()   except +

    ctypedef vector[Comment] CommentList

    
    cdef cppclass DigInEv:
        uint32_t    timestamp
        Source      source
        uint32_t    data
        DigInEv()   except +

    ctypedef vector[DigInEv] DigInEvList

cdef extern from "sre/UnitHist.h" namespace "cc::sre":
    cdef cppclass UnitHist[T, C=*]:
        ctypedef int size_type
        pass

cdef extern from "sre/SpikeRateEstimator.h" namespace "cc::sre":
    cdef cppclass SpikeRateEstimator:
        SpikeRateEstimator() except +
        ctypedef UnitHist[unsigned int] unit_hist_type
        ctypedef vector[unit_hist_type] unit_hist_container_type
        ctypedef int size_type
        ctypedef pair[size_type, size_type] ch_un_type
        ctypedef vector[ch_un_type] ch_un_list_type
        
        #cppclass ChUnHash:
        #    int operator()(const ch_un_type & p) const
        
        #ctypedef unordered_map[ch_un_type, size_type, ChUnHash] ch_un_map_type
        
        ctypedef double rate_type
    
        ctypedef pair[uint32_t, vector[rate_type]] rate_list_item_type
        ctypedef vector[rate_list_item_type] rate_list_type
        
        void processSpikes()
        
        void setChannelUnitList(const ch_un_list_type & ch_un_list)
        void fillChannelUnitList(const size_type n_channels, const size_type n_units)
        ch_un_list_type getChannelUnitList() const


cdef extern from "sbpe/SpikeBandPowerEstimator.h" namespace "cc::sbpe":
    cdef cppclass SpikeBandPowerEstimator:
        SpikeBandPowerEstimator() except +
        ctypedef size_t size_type
        ctypedef double sbp_type
        ctypedef vector[sbp_type] sbp_item_type
        ctypedef pair[uint32_t, sbp_item_type] sbp_ring_buffer_item_type
        ctypedef vector[sbp_ring_buffer_item_type] sbp_list_type
        
        ctypedef size_type ch_type
        ctypedef vector[ch_type] ch_list_type
        
        ctypedef ac.vec iir_filt_coeff_type
        
        

cdef extern from "CereConn.h" namespace "cc::CereConn":
    cdef cppclass State:
        pass
    cdef cppclass ReturnCode:
        pass

ctypedef uint32_t Timestamp_t

cdef extern from "CereConn.h" namespace "cc":
    cdef cppclass CereConn:
    
        CereConn() except +
        CereConn(uint32_t) except +
        CereConn(uint32_t, bool) except +
        CereConn(uint32_t, bool, bool) except +
        void ccOpen()
        void ccRecord()
        void ccIdle()
        void ccClose()
        CereConn.State getState()
        CereConn.ReturnCode getSampleGroupData(SampleGroupList &sgl, uint32_t &cbTime)
        CereConn.ReturnCode getCbTime(uint32_t &cbTime)
        void printBufferState()
        CereConn.ReturnCode sendComment(const string comment, const uint32_t rgba, const uint8_t charset)
        CereConn.ReturnCode getCommentData(CommentList &clist, uint32_t &cbTime)
        CereConn.ReturnCode getCommentData(CommentList &clist, uint32_t &cbTime, size_t &overflowCount)

        CereConn.ReturnCode getDigInEvData(DigInEvList &dielist, uint32_t &cbTime)
        CereConn.ReturnCode getDigInEvData(DigInEvList &dielist, uint32_t &cbTime, size_t &overflowCount)

        void setSampleGroup(size_t group, bool state)
        bool getSampleGroup(size_t group)
        
        unsigned int getSampleGroupPeriod(const size_t group)
        unsigned int getSampleGroupFrequency(const size_t group)

        bool setCARChannels(const size_t group, const ChannelNumList_t &carChannelList)
        void disableCAR(const size_t group)
        
        
        void spikeRateEstimatorStart()
        void spikeRateEstimatorStop()
        CereConn.ReturnCode getSpikeRateData(SpikeRateEstimator.rate_list_type &, uint32_t &cbTime)
        CereConn.ReturnCode getSpikeRateData(SpikeRateEstimator.rate_list_type &, uint32_t &cbTime, size_t &overflowCount)
        # access to spike rate estimator's channel+unit mapping
        void setSpikeRateEstimatorChannelUnitList(const SpikeRateEstimator.ch_un_list_type & ch_un_list)
        void fillSpikeRateEstimatorChannelUnitList(const SpikeRateEstimator.size_type n_channels, const SpikeRateEstimator.size_type n_units)
        SpikeRateEstimator.ch_un_list_type getSpikeRateEstimatorChannelUnitList() const
        
        void setSpikeRateEstimatorLoopIntervalMS(const unsigned int & new_interval)
        unsigned int getSpikeRateEstimatorLoopIntervalMS() const
    
        void spikeRateEstimatorId()
        void spikeRateEstimatorBoxcar(const size_t length)
        void spikeRateEstimatorExp(const double decay_factor, const size_t max_length)
    
        void spikeBandPowerEstimatorStart()
        void spikeBandPowerEstimatorStop()
        void setSpikeBandPowerEstimatorChannelList(const SpikeBandPowerEstimator.ch_list_type & ch_list)
        void fillSpikeBandPowerEstimatorChannelList(const SpikeBandPowerEstimator.size_type n_channels)
        SpikeBandPowerEstimator.ch_list_type getSpikeBandPowerEstimatorChannelList() const
        
        void setSpikeBandPowerEstimatorLoopIntervalMS(const unsigned int & new_interval)
        unsigned int getSpikeBandPowerEstimatorLoopIntervalMS() const
        
        void setSpikeBandPowerEstimatorAcceptedSampleGroup(const size_t group)
        size_t getSpikeBandPowerEstimatorAcceptedSampleGroup() const
        
        void setSpikeBandPowerEstimatorIIRFilterNSamples(const size_t n_samples)
        size_t getSpikeBandPowerEstimatorIIRFilterNSamples()
        
        void setSpikeBandPowerEstimatorAvgNBins(const size_t n_bins)
        size_t getSpikeBandPowerEstimatorAvgNBins() const
        
        
        CereConn.ReturnCode getSpikeBandPowerData(SpikeBandPowerEstimator.sbp_list_type &sbpl, uint32_t &cbTime)
        CereConn.ReturnCode getSpikeBandPowerData(SpikeBandPowerEstimator.sbp_list_type &sbpl, uint32_t &cbTime, size_t &overflowCount)

        void setSpikeBandPowerEstimatorIIRFilterCoeffs(SpikeBandPowerEstimator.iir_filt_coeff_type b, SpikeBandPowerEstimator.iir_filt_coeff_type a)
        pair[SpikeBandPowerEstimator.iir_filt_coeff_type, SpikeBandPowerEstimator.iir_filt_coeff_type] getSpikeBandPowerEstimatorIIRFilterCoeffs()
        
        
        #void bindSpikeRateEstimatorCallback(const SpikeRateEstimator.rate_callback_type& rcb)
        
        

