#!/usr/bin/env python3
## @package test_01
# Comment Test 01
#
# @author Jonas Zimmermann
#
# This test will open a connection to an NSP, and it will send and retrieve a comment.

import time

import cere_conn as cc

ck = cc.CereConn()

ck.send_open()

t = time.time()
while ck.get_state() != cc.ccS_Idle:
    time.sleep(0.005)
print("It took {:5.3f}s to open CereConn\n".format(time.time() - t))

ck.send_record()

t = time.time()
while ck.get_state() != cc.ccS_Recording:
    time.sleep(0.2)
print("It took {:5.3f}s to start CereConn recording\n".format(time.time() - t))


n_comments = 50

start_time = time.time()

round_trip_times = []
round_trip_times_abs = []

ck.get_comment_data()
for i in range(n_comments):
    cb_time = ck.get_cb_time()['ts']
    ck.send_comment("Test {} at tick {}".format(i+1, cb_time), cb_time)
    cd = ck.get_comment_data()
    while cb_time not in [c['data'] for c in cd['comments'] ]:
        time.sleep(0.001)
        cd = ck.get_comment_data()
        
    com = [com for com in cd['comments'] if com['data'] == cb_time]
    round_trip_times.append(com[0]['ts'] - com[0]['data'])
    round_trip_times_abs.append(cd['ts'] - com[0]['data'])

    
print("Average round trip times for {} comments: {} ticks, or {:5.1f}ms (min {:5.1f}, max {:5.1f})".format(n_comments, sum(round_trip_times)/len(round_trip_times), sum(round_trip_times)/len(round_trip_times) / 30.0, min(round_trip_times)/30.0, max(round_trip_times)/30.0 ))
print("Average round trip times for {} comments until processed here: {} ticks, or {:5.1f}ms (min {:5.1f}, max {:5.1f})".format(n_comments, sum(round_trip_times_abs)/len(round_trip_times_abs), sum(round_trip_times_abs)/len(round_trip_times_abs) / 30.0, min(round_trip_times_abs)/30.0, max(round_trip_times_abs)/30.0 ))


ck.send_close()
while ck.get_state() != cc.ccS_Closed:
    time.sleep(0.05)



