#!/usr/bin/env python3
## @package test_01
# Continuous Test 01
#
# @author Jonas Zimmermann
#
# This test will open a connection to an NSP, and it will retrieve continuous data.

import time
import cere_conn as cc

# get CereConn object
ck = cc.CereConn()
ck.send_open()

# Wait until connection is established
t = time.time()
while ck.get_state() != cc.ccS_Idle:
    time.sleep(0.005)
print("It took {:5.3f}s to open CereConn\n".format(time.time() - t))

# start recording
ck.send_record()

t = time.time()
while ck.get_state() != cc.ccS_Recording:
    time.sleep(0.05)
print("It took {:5.3f}s to start CereConn recording\n".format(time.time() - t))

# at least with NPlayServer, we need to wait for things to settle
time.sleep(.5)


# request sample group 2
gids = [2, 4, 5, 6]
ignore_gids = [2, 5]
for gid in gids:
    ck.set_sample_group_state(gid, True)


bin_width = 0.005

run_time = 5.0

for g in ignore_gids:
    ck.set_sample_group_state(g, False)
cd = ck.get_cont_data()

# keep stats...
lasttime = time.time()
starttime = lasttime
times = []
samples = {}

# run data acquisition for a while
while (time.time() - starttime < run_time):
    cd = ck.get_cont_data(sample_groups=gids)
    # store number of samples
    for gid in gids:
        samples.setdefault(gid, []).append(cd['sample_groups'][gid]['data'].shape[0])
    ctime = time.time()
    times.append( ctime - lasttime)
    lasttime = ctime
    time.sleep(bin_width)

ck.send_close()


print("Some stats: Bin width: {:4.3f}s".format(bin_width))
for gid in gids:
    print("Sample group {} has these channels: {}".format(gid, cd['sample_groups'][gid]['ch']))
    print("\tReceived {} samples in {:5.3f}s. That is {:5.1f} samples / s.".format(sum(samples[gid]), sum(times), sum(samples[gid])/sum(times)))
    print("\tThat's {:8.1f} samples per bin on average".format(sum(samples[gid])/len(samples[gid])))

while ck.get_state() != cc.ccS_Closed:
    time.sleep(0.05)
