#include <algorithm>
#include <chrono>
#include <cmath>
#include <cstdio>
#include "SdkApp.h"
#include "cbsdk.h"
#include "cbsdk_debug.h"

#define _snprintf snprintf

// The sdk instances
SdkApp * g_app[cbMAXOPEN] = {NULL};


// Author & Date: Ehsan Azar       29 April 2012
/// Sdk app base constructor
SdkApp::SdkApp()
    : m_bDone(false), m_instInfo(0), m_nInstance(0)
    , m_bInitialized(false), m_lastCbErr(cbRESULT_OK)
    , m_bPacketsEvent(false), m_bPacketsCmt(false), m_bPacketsTrack(false)
    , m_uCbsdkTime(0)
{
    for (int i = 0; i < CBSDKCALLBACK_COUNT; ++i)
    {
        m_pCallback[i] = 0;
        m_pCallbackParams[i] = 0;
        m_pLateCallback[i] = 0;
        m_pLateCallbackParams[i] = 0;
    }
}

// Author & Date: Ehsan Azar       29 April 2012
/// Sdk app base destructor
SdkApp::~SdkApp()
{
    // Close networking
    Close();
}

// Author & Date: Ehsan Azar       29 April 2012
/** Open sdk application, network and Qt message loop.
*
* @param[in]	nInstance		instance ID
* @param[in]	nInPort			Client port number
* @param[in]	nOutPort		Instrument port number
* @param[in]	szInIP			Client IPv4 address
* @param[in]	szOutIP			Instrument IPv4 address
* @param[in]	nRecBufSize		reserved
*/
void SdkApp::Open(uint32_t nInstance, int nInPort, int nOutPort, LPCSTR szInIP, LPCSTR szOutIP, int nRecBufSize)
{
    // clear lats library error
    m_lastCbErr = cbRESULT_OK;
    // Close networking thread if already running
    Close();
    // One-time initialization
    if (!m_bInitialized)
    {
        m_bInitialized = true;
    }
    // instance id and connection details are persistent in the process
    m_nInstance = nInstance;

    m_instInfo = cbINSTINFO_LOCAL;
    // Restart networking thread
    startThread();
}

void SdkApp::Close()
{
    // Signal thread to finish
    m_bDone = true;
    if (m_thread.joinable())
    {
        m_thread.join();
    }
}

void SdkApp::threadFun()
{
    using tick_interval = std::chrono::duration<long long, std::ratio<1, 30000>>;

    cbPKT_GROUP my_pkt_001({0,0,1,1,{0}});
    my_pkt_001.data[0] = 1;

    cbPKT_GROUP my_pkt_002({0,0,2,1,{0}});
    my_pkt_002.data[0] = 2;
    my_pkt_002.data[1] = 3;

    cbPKT_GROUP my_pkt_003({0,0,3,2,{0}});
    my_pkt_003.data[0] = 4;
    my_pkt_003.data[1] = 5;
    my_pkt_003.data[2] = 6;

    cbPKT_GROUP my_pkt_004({0,0,4,1,{0}});
    my_pkt_004.data[0] = 7;
    my_pkt_004.data[1] = 8;

    cbPKT_GROUP my_pkt_005({0,0,5,1,{0}});
    my_pkt_005.data[0] = 9;

    cbPKT_GROUP my_pkt_006({0,0,6, cbNUM_FE_CHANS/2,{0}});
    for (size_t i = 0; i < cbNUM_FE_CHANS; ++i)
    {
        my_pkt_006.data[i] = i + 1;
    }

    cbPKT_SPK my_spk01({0, 1, 0});
    cbPKT_SPK my_spk02({0, 2, 1});
    
    cbPKT_DINP my_pkt_digin({0, MAX_CHANS_DIGITAL_IN, 0, 1, 42});
    cbPKT_DINP my_pkt_serial({0, MAX_CHANS_SERIAL, 0, 1, 82});
    
    size_t time_ctr = 0;
    auto start = std::chrono::high_resolution_clock::now();

    while (! m_bDone)
    {
        if (((time_ctr+1) % 15223) == 0)
        {
            my_pkt_digin.time = time_ctr;
            ++my_pkt_digin.data[0];
            ProcessIncomingPacket((cbPKT_GENERIC *)&my_pkt_digin);
        }
        if (((time_ctr+1) % 8923) == 0)
        {
            my_pkt_serial.time = time_ctr;
            ++my_pkt_serial.data[0];
            
            ProcessIncomingPacket((cbPKT_GENERIC *)&my_pkt_serial);
        }
        
        
        if (((time_ctr+1) % 1304) == 0)
        {
            my_spk01.time = time_ctr;
            ProcessIncomingPacket((cbPKT_GENERIC *)&my_spk01);
        }
        if (((time_ctr+1) % 547) == 0)
        {
            my_spk02.time = time_ctr;
            ProcessIncomingPacket((cbPKT_GENERIC *)&my_spk02);
        }
        
        if ((time_ctr % 60) == 0)
        {
            my_pkt_001.time = time_ctr;
            my_pkt_001.data[0] = 1 + (time_ctr % 2);
            ProcessIncomingPacket((cbPKT_GENERIC *)&my_pkt_001);
        }
        if ((time_ctr % 30) == 0)
        {
            my_pkt_002.time = time_ctr;
            my_pkt_002.data[0] = 2 + (time_ctr % 3);
            ProcessIncomingPacket((cbPKT_GENERIC *)&my_pkt_002);
        }

        if ((time_ctr % 15) == 0)
        {
            my_pkt_003.time = time_ctr;
            my_pkt_003.data[0] = 4 + (time_ctr % 4);
            ProcessIncomingPacket((cbPKT_GENERIC *)&my_pkt_003);
        }

        if ((time_ctr % 3) == 0)
        {
            my_pkt_004.time = time_ctr;
            my_pkt_004.data[1] = 500 * sin(time_ctr * 3.1415927 * 2.0 * 1500.0 / 1e4  );
            ProcessIncomingPacket((cbPKT_GENERIC *)&my_pkt_004);
        }
        if ((time_ctr % 1) == 0)
        {
            my_pkt_005.time = time_ctr;
            my_pkt_005.data[0] = 9 + (time_ctr % 6);
            ProcessIncomingPacket((cbPKT_GENERIC *)&my_pkt_005);

            my_pkt_006.time = time_ctr;
            my_pkt_006.data[0] = 1000 * sin(time_ctr * 3.1415927 * 2.0 * 1.0 / 3e4  );
            my_pkt_006.data[1] = 1000 * sin(time_ctr * 3.1415927 * 2.0 * 5.0 / 3e4  );
            my_pkt_006.data[2] = 1000 * sin(time_ctr * 3.1415927 * 2.0 * 10.0 / 3e4  );
            my_pkt_006.data[3] = 1000 * sin(time_ctr * 3.1415927 * 2.0 * 50.0 / 3e4  );
            my_pkt_006.data[4] = 1000 * sin(time_ctr * 3.1415927 * 2.0 * 100.0 / 3e4  );
            my_pkt_006.data[5] = 1000 * sin(time_ctr * 3.1415927 * 2.0 * 500.0 / 3e4  );
            my_pkt_006.data[6] = 1000 * sin(time_ctr * 3.1415927 * 2.0 * 1000.0 / 3e4  );
            my_pkt_006.data[7] = 1000 * sin(time_ctr * 3.1415927 * 2.0 * 5000.0 / 3e4  );
            my_pkt_006.data[8] = 1000 * sin(time_ctr * 3.1415927 * 2.0 * 10000.0 / 3e4  );
            my_pkt_006.data[9] = 1000 * sin(time_ctr * 3.1415927 * 2.0 * 15000.0 / 3e4  );
            ProcessIncomingPacket((cbPKT_GENERIC *)&my_pkt_006);
        }

        ++time_ctr;

        // This loop ensures that one tick is ~ 1/30 ms. Comment out to produce samples as fast as possible.
#ifndef USE_CERELINK_STUB_PERFORMANCE
        while (! m_bDone && (std::chrono::high_resolution_clock::now() < start + tick_interval(time_ctr)))
        {
            std::this_thread::sleep_for(std::chrono::nanoseconds(50));
        }
#endif
    }
}
void SdkApp::startThread()
{
    if (m_thread.joinable())
    {
        m_bDone = true;
        m_thread.join();
    }
    m_bDone = false;
    m_thread = std::thread(&SdkApp::threadFun, this);

}


// Author & Date:   Ehsan Azar     16 May 2012
/** Late binding of callback function when needed.
*
* Make sure this is called before any callback invokation.
*
* @param[in] callbacktype the calback type to bind
*/
void SdkApp::LateBindCallback(cbSdkCallbackType callbacktype)
{
    if (m_pCallback[callbacktype] != m_pLateCallback[callbacktype])
    {
        std::lock_guard<std::mutex> lock(m_lockCallback);
        m_pLateCallback[callbacktype] = m_pCallback[callbacktype];
        m_pLateCallbackParams[callbacktype] = m_pCallbackParams[callbacktype];
    }
}



// Author & Date:   Ehsan Azar     23 Feb 2011
/** Open cbsdk library.
*
* @param[in] nInstance		Instance ID
* @param[in] conType		The requested connection type to open
* @param[in] con			Connection details

* \n This function returns the error code
*/
cbSdkResult SdkApp::SdkOpen(uint32_t nInstance, cbSdkConnectionType conType, cbSdkConnection con)
{
    // check if the library is already open
    // Some sanity checks
    if (con.szInIP == NULL || con.szInIP[0] == 0)
    {
#ifdef WIN32
        con.szInIP = cbNET_UDP_ADDR_INST;
#elif __APPLE__
        con.szInIP = "255.255.255.255";
#else
        // On Linux bind to bcast
        con.szInIP = cbNET_UDP_ADDR_BCAST;
#endif
    }
    if (con.szOutIP == NULL || con.szOutIP[0] == 0)
        con.szOutIP = cbNET_UDP_ADDR_CNT;

    // Unregister all callbacks
    {
        std::lock_guard<std::mutex> lock(m_lockCallback);
        for (int i = 0; i < CBSDKCALLBACK_COUNT; ++i)
        {
            m_pCallbackParams[i] = NULL;
            m_pCallback[i] = NULL;
        }
    }

    // open the library and verify that the central application is there
    conType = CBSDKCONNECTION_UDP; // Now try UDP

    if (conType == CBSDKCONNECTION_UDP)
    {
        Open(nInstance, con.nInPort, con.nOutPort, con.szInIP, con.szOutIP, con.nRecBufSize);
    }
    else if (conType == CBSDKCONNECTION_CENTRAL)
    {
        Open(nInstance);
    }
    else
    {
        return CBSDKRESULT_NOTIMPLEMENTED;
    }

    // Wait for (dis)connection to happen
        if (m_instInfo == 0)
        {
            cbSdkResult sdkres = CBSDKRESULT_UNKNOWN;
            // Try to make sense of the error
            switch (GetLastCbErr())
            {
            case cbRESULT_NOCENTRALAPP:
                sdkres = CBSDKRESULT_ERROPENCENTRAL;
                break;
            case cbRESULT_SOCKERR:
                sdkres = CBSDKRESULT_ERROPENUDPPORT;
                break;
            case cbRESULT_SOCKOPTERR:
                sdkres = CBSDKRESULT_OPTERRUDP;
                break;
            case cbRESULT_SOCKMEMERR:
                sdkres = CBSDKRESULT_MEMERRUDP;
                break;
            case cbRESULT_INSTINVALID:
                sdkres = CBSDKRESULT_INVALIDINST;
                break;
            case cbRESULT_SOCKBIND:
                sdkres = CBSDKRESULT_ERROPENUDP;
                break;
            case cbRESULT_LIBINITERROR:
            case cbRESULT_EVSIGERR:
                sdkres = CBSDKRESULT_ERRINIT;
                break;
            case cbRESULT_SYSLOCK:
                sdkres = CBSDKRESULT_BUSY;
                break;
            case cbRESULT_BUFRECALLOCERR:
            case cbRESULT_BUFGXMTALLOCERR:
            case cbRESULT_BUFLXMTALLOCERR:
            case cbRESULT_BUFCFGALLOCERR:
            case cbRESULT_BUFPCSTATALLOCERR:
            case cbRESULT_BUFSPKALLOCERR:
                sdkres = CBSDKRESULT_ERRMEMORY;
                break;
            case cbRESULT_INSTOUTDATED:
                sdkres = CBSDKRESULT_INSTOUTDATED;
                break;
            case cbRESULT_LIBOUTDATED:
                sdkres = CBSDKRESULT_LIBOUTDATED;
                break;
            case cbRESULT_OK:
                sdkres = CBSDKRESULT_ERROFFLINE;
                break;
            }
            return sdkres;
        }

    return CBSDKRESULT_SUCCESS;
}


CBSDKAPI    cbSdkResult cbSdkOpen(uint32_t nInstance,
                                  cbSdkConnectionType conType,
                                  cbSdkConnection con)
{
    // check if the library is already open
    if (conType < 0 || conType >= CBSDKCONNECTION_CLOSED)
        return CBSDKRESULT_INVALIDPARAM;
    if (nInstance >= cbMAXOPEN)
        return CBSDKRESULT_INVALIDPARAM;
    if (g_app[nInstance] == NULL)
    {
        try {
            g_app[nInstance] = new SdkApp();
        } catch (...) {
            g_app[nInstance] = NULL;
        }
    }
    if (g_app[nInstance] == NULL)
        return CBSDKRESULT_ERRMEMORY;

    return g_app[nInstance]->SdkOpen(nInstance, conType, con);
}


// Author & Date:   Ehsan Azar     24 Feb 2011
/** Close cbsdk library.
*
* This function returns the error code
*/
cbSdkResult SdkApp::SdkClose()
{
    cbSdkResult res = CBSDKRESULT_SUCCESS;
    if (m_instInfo == 0)
        res = CBSDKRESULT_WARNCLOSED;

    // clear the data cache if it exists
    // first disable writing to the cache

    // Unregister all callbacks
    {
        std::lock_guard<std::mutex> lock(m_lockCallback);
        for (int i = 0; i < CBSDKCALLBACK_COUNT; ++i)
        {
            m_pCallbackParams[i] = NULL;
            m_pCallback[i] = NULL;
        }
    }
    // Close the app
    Close();

    return res;
}

CBSDKAPI    cbSdkResult cbSdkClose(uint32_t nInstance)
{
    if (nInstance >= cbMAXOPEN)
        return CBSDKRESULT_INVALIDPARAM;
    if (g_app[nInstance] == NULL)
        return CBSDKRESULT_WARNCLOSED;

    cbSdkResult res = g_app[nInstance]->SdkClose();

    // Delete this instance
    delete g_app[nInstance];
    g_app[nInstance] = NULL;
    return res;
}

cbSdkResult SdkApp::SdkGetTime(uint32_t * cbtime)
{
    if (m_instInfo == 0)
        return CBSDKRESULT_CLOSED;

    *cbtime = m_uCbsdkTime;

    return CBSDKRESULT_SUCCESS;
}

CBSDKAPI    cbSdkResult cbSdkGetTime(uint32_t nInstance, uint32_t * cbtime)
{
    if (cbtime == NULL)
        return CBSDKRESULT_NULLPTR;
    if (nInstance >= cbMAXOPEN)
        return CBSDKRESULT_INVALIDPARAM;
    if (g_app[nInstance] == NULL)
        return CBSDKRESULT_CLOSED;

    return g_app[nInstance]->SdkGetTime(cbtime);
}

cbSdkResult SdkApp::SdkRegisterCallback(cbSdkCallbackType callbacktype, cbSdkCallback pCallbackFn, void * pCallbackData)
{
    if (m_instInfo == 0)
        return CBSDKRESULT_CLOSED;
    if (m_pCallback[callbacktype]) // Already registered
        return CBSDKRESULT_CALLBACKREGFAILED;
    {
        std::lock_guard<std::mutex> lock(m_lockCallback);
        m_pCallbackParams[callbacktype] = pCallbackData;
        m_pCallback[callbacktype] = pCallbackFn;
    }
    return CBSDKRESULT_SUCCESS;
}

CBSDKAPI    cbSdkResult cbSdkRegisterCallback(uint32_t nInstance, cbSdkCallbackType callbacktype, cbSdkCallback pCallbackFn, void* pCallbackData)
{
    if (!pCallbackFn)
        return CBSDKRESULT_NULLPTR;
    if (callbacktype >= CBSDKCALLBACK_COUNT)
        return CBSDKRESULT_INVALIDCALLBACKTYPE;
    if (nInstance >= cbMAXOPEN)
        return CBSDKRESULT_INVALIDPARAM;
    if (g_app[nInstance] == NULL)
        return CBSDKRESULT_CLOSED;

    return g_app[nInstance]->SdkRegisterCallback(callbacktype, pCallbackFn, pCallbackData);
}

cbSdkResult SdkApp::SdkUnRegisterCallback(cbSdkCallbackType callbacktype)
{
    if (m_instInfo == 0)
        return CBSDKRESULT_CLOSED;
    if (!m_pCallback[callbacktype]) // Already unregistered
        return CBSDKRESULT_CALLBACKREGFAILED;

    std::lock_guard<std::mutex> lock(m_lockCallback);
    m_pCallback[callbacktype] = NULL;
    m_pCallbackParams[callbacktype] = NULL;
    return CBSDKRESULT_SUCCESS;
}

CBSDKAPI    cbSdkResult cbSdkUnRegisterCallback(uint32_t nInstance, cbSdkCallbackType callbacktype)
{
    if (callbacktype >= CBSDKCALLBACK_COUNT)
        return CBSDKRESULT_INVALIDCALLBACKTYPE;
    if (nInstance >= cbMAXOPEN)
        return CBSDKRESULT_INVALIDPARAM;
    if (g_app[nInstance] == NULL)
        return CBSDKRESULT_CLOSED;

    return g_app[nInstance]->SdkUnRegisterCallback(callbacktype);
}

CBSDKAPI    cbSdkResult cbSdkSetComment(uint32_t nInstance, uint32_t rgba, uint8_t charset, const char * comment)
{
    if (nInstance >= cbMAXOPEN)
        return CBSDKRESULT_INVALIDPARAM;
    if (g_app[nInstance] == NULL)
        return CBSDKRESULT_CLOSED;
    uint32_t ts;
    cbSdkGetTime(nInstance, &ts);
    cbPKT_COMMENT pkt_cmt {ts, 0x8000, cbPKTTYPE_COMMENTREP, cbPKTDLEN_COMMENT, {charset, cbCOMMENT_FLAG_RGBA, {0,0}}, rgba,""};
    strncpy(pkt_cmt.comment, comment ,cbMAX_COMMENT);
    g_app[nInstance]->ProcessIncomingPacket((cbPKT_GENERIC *)&pkt_cmt);
    return CBSDKRESULT_SUCCESS;
}

CBSDKAPI    cbSdkResult cbSdkGetSampleGroupInfo(uint32_t nInstance, uint32_t proc, uint32_t group, char *label, uint32_t *period, uint32_t *length)
{
    uint32_t _length = 0;
    uint32_t _period = 0;
    switch (group)
    {
        case 1:
            _length = 1;
            _period = 60;
            break;
        case 2:
            _length = 2;
            _period = 30;
            break;
        case 3:
            _length = 3;
            _period = 15;
            break;
        case 4:
            _length = 2;
            _period = 3;
            break;
        case 5:
            _length = 1;
            _period = 1;
            break;
        case 6:
            _length = cbNUM_FE_CHANS;
            _period = 1;
            break;
        default:
            _length = 0;
            _period = 1;
            break;
    }
    if (length != nullptr)
    {
        *length = _length;
    }
    if (period != nullptr)
    {
        *period = _period;
    }

    return CBSDKRESULT_SUCCESS;
}

CBSDKAPI    cbSdkResult cbSdkGetSampleGroupList(uint32_t nInstance, uint32_t proc, uint32_t group, uint32_t *length, uint32_t *list)
{
    uint32_t _length = 0;
    uint32_t _list[cbNUM_ANALOG_CHANS] = {0};

    switch (group)
    {
        case 1:
            _length = 1;
            _list[0] = 1;
            break;
        case 2:
            _length = 2;
            _list[0] = 2;
            _list[1] = 3;
            break;
        case 3:
            _length = 3;
            _list[0] = 4;
            _list[1] = 5;
            _list[2] = 6;
            break;
        case 4:
            _length = 2;
            _list[0] = 7;
            _list[1] = 8;
            break;
        case 5:
            _length = 1;
            _list[0] = 9;
            break;
        case 6:
            _length = cbNUM_FE_CHANS;
            for (size_t i = 0; i < cbNUM_FE_CHANS; ++ i)
            {
                _list[i] = i + 1;
            }
            break;
        default:
            _length = 0;
            break;
    }
    if (length != nullptr)
    {
        *length = _length;
    }

    if (list != nullptr)
    {
        std::copy(_list, _list + _length, list);
    }


    return CBSDKRESULT_SUCCESS;
}




// Author & Date: Ehsan Azar       29 April 2012
/// Proxy for all incomming packets
void SdkApp::ProcessIncomingPacket(const cbPKT_GENERIC * const pPkt)
{
    // This is a hot code path, and crosses the shared library
    //  we want as minimal locking as possible
    //  we also want to reduce deadlock if someone calls unregister within the callback itself
    //  thus we late bind callback when a change is noticed and only when needed,
    //   then use the late bound callback function
    //  As a rule of thumb, no locks should be active before any callback is called
    // This callback type needs to be bound only once
    LateBindCallback(CBSDKCALLBACK_ALL);

    // check for configuration class packets
    if (pPkt->chid & cbPKTCHAN_CONFIGURATION)
    {
        // Check for configuration packets
        if (pPkt->chid == cbPKTCHAN_CONFIGURATION)
        {
            if (pPkt->type == cbPKTTYPE_SYSHEARTBEAT)
            {
                if (m_pLateCallback[CBSDKCALLBACK_ALL])
                    m_pLateCallback[CBSDKCALLBACK_ALL](m_nInstance, cbSdkPkt_SYSHEARTBEAT, pPkt, m_pLateCallbackParams[CBSDKCALLBACK_ALL]);
                // Late bind before usage
                LateBindCallback(CBSDKCALLBACK_SYSHEARTBEAT);
                if (m_pLateCallback[CBSDKCALLBACK_SYSHEARTBEAT])
                    m_pLateCallback[CBSDKCALLBACK_SYSHEARTBEAT](m_nInstance, cbSdkPkt_SYSHEARTBEAT, pPkt, m_pLateCallbackParams[CBSDKCALLBACK_SYSHEARTBEAT]);
            }
            else if (pPkt->type == cbPKTTYPE_REPIMPEDANCE)
            {
                if (m_pLateCallback[CBSDKCALLBACK_ALL])
                    m_pLateCallback[CBSDKCALLBACK_ALL](m_nInstance, cbSdkPkt_IMPEDANCE, pPkt, m_pLateCallbackParams[CBSDKCALLBACK_ALL]);
                // Late bind before usage
                LateBindCallback(CBSDKCALLBACK_IMPEDENCE);
                if (m_pLateCallback[CBSDKCALLBACK_IMPEDENCE])
                    m_pLateCallback[CBSDKCALLBACK_IMPEDENCE](m_nInstance, cbSdkPkt_IMPEDANCE, pPkt, m_pLateCallbackParams[CBSDKCALLBACK_IMPEDENCE]);
            }
            else if ((pPkt->type & 0xF0) == cbPKTTYPE_CHANREP)
            {
                if (m_pLateCallback[CBSDKCALLBACK_ALL])
                    m_pLateCallback[CBSDKCALLBACK_ALL](m_nInstance, cbSdkPkt_CHANINFO, pPkt, m_pLateCallbackParams[CBSDKCALLBACK_ALL]);
                // Late bind before usage
                LateBindCallback(CBSDKCALLBACK_CHANINFO);
                if (m_pLateCallback[CBSDKCALLBACK_CHANINFO])
                    m_pLateCallback[CBSDKCALLBACK_CHANINFO](m_nInstance, cbSdkPkt_CHANINFO, pPkt, m_pLateCallbackParams[CBSDKCALLBACK_CHANINFO]);
            }
            else if (pPkt->type == cbPKTTYPE_NMREP)
            {
                if (m_pLateCallback[CBSDKCALLBACK_ALL])
                    m_pLateCallback[CBSDKCALLBACK_ALL](m_nInstance, cbSdkPkt_NM, pPkt, m_pLateCallbackParams[CBSDKCALLBACK_ALL]);
                // Late bind before usage
                LateBindCallback(CBSDKCALLBACK_NM);
                if (m_pLateCallback[CBSDKCALLBACK_NM])
                    m_pLateCallback[CBSDKCALLBACK_NM](m_nInstance, cbSdkPkt_NM, pPkt, m_pLateCallbackParams[CBSDKCALLBACK_NM]);
            }
            else if (pPkt->type == cbPKTTYPE_GROUPREP)
            {
                if (m_pLateCallback[CBSDKCALLBACK_ALL])
                    m_pLateCallback[CBSDKCALLBACK_ALL](m_nInstance, cbSdkPkt_GROUPINFO, pPkt, m_pLateCallbackParams[CBSDKCALLBACK_ALL]);
                // Late bind before usage
                LateBindCallback(CBSDKCALLBACK_GROUPINFO);
                if (m_pLateCallback[CBSDKCALLBACK_GROUPINFO])
                    m_pLateCallback[CBSDKCALLBACK_GROUPINFO](m_nInstance, cbSdkPkt_GROUPINFO, pPkt, m_pLateCallbackParams[CBSDKCALLBACK_GROUPINFO]);
            }
            else if (pPkt->type == cbPKTTYPE_COMMENTREP)
            {
                if (m_pLateCallback[CBSDKCALLBACK_ALL])
                    m_pLateCallback[CBSDKCALLBACK_ALL](m_nInstance, cbSdkPkt_COMMENT, pPkt, m_pLateCallbackParams[CBSDKCALLBACK_ALL]);
                // Late bind before usage
                LateBindCallback(CBSDKCALLBACK_COMMENT);
                if (m_pLateCallback[CBSDKCALLBACK_COMMENT])
                    m_pLateCallback[CBSDKCALLBACK_COMMENT](m_nInstance, cbSdkPkt_COMMENT, pPkt, m_pLateCallbackParams[CBSDKCALLBACK_COMMENT]);
            }
            else if (pPkt->type == cbPKTTYPE_REPFILECFG)
            {
                if (m_pLateCallback[CBSDKCALLBACK_ALL])
                    m_pLateCallback[CBSDKCALLBACK_ALL](m_nInstance, cbSdkPkt_FILECFG, pPkt, m_pLateCallbackParams[CBSDKCALLBACK_ALL]);
                // Late bind before usage
                LateBindCallback(CBSDKCALLBACK_FILECFG);
                if (m_pLateCallback[CBSDKCALLBACK_FILECFG])
                    m_pLateCallback[CBSDKCALLBACK_FILECFG](m_nInstance, cbSdkPkt_FILECFG, pPkt, m_pLateCallbackParams[CBSDKCALLBACK_COMMENT]);
            }
            else if (pPkt->type == cbPKTTYPE_REPPOLL)
            {
                // The callee should check flags to find if it is a response to poll, and do accordingly
                if (m_pLateCallback[CBSDKCALLBACK_ALL])
                    m_pLateCallback[CBSDKCALLBACK_ALL](m_nInstance, cbSdkPkt_POLL, pPkt, m_pLateCallbackParams[CBSDKCALLBACK_ALL]);
                // Late bind before usage
                LateBindCallback(CBSDKCALLBACK_POLL);
                if (m_pLateCallback[CBSDKCALLBACK_POLL])
                    m_pLateCallback[CBSDKCALLBACK_POLL](m_nInstance, cbSdkPkt_POLL, pPkt, m_pLateCallbackParams[CBSDKCALLBACK_POLL]);
            }
            else if (pPkt->type == cbPKTTYPE_VIDEOTRACKREP)
            {
                if (m_pLateCallback[CBSDKCALLBACK_ALL])
                    m_pLateCallback[CBSDKCALLBACK_ALL](m_nInstance, cbSdkPkt_TRACKING, pPkt, m_pLateCallbackParams[CBSDKCALLBACK_ALL]);
                // Late bind before usage
                LateBindCallback(CBSDKCALLBACK_TRACKING);
                if (m_pLateCallback[CBSDKCALLBACK_TRACKING])
                    m_pLateCallback[CBSDKCALLBACK_TRACKING](m_nInstance, cbSdkPkt_TRACKING, pPkt, m_pLateCallbackParams[CBSDKCALLBACK_TRACKING]);
                // Fillout trial if setup
                // OnPktTrack(reinterpret_cast<const cbPKT_VIDEOTRACK*>(pPkt));
            }
            else if (pPkt->type == cbPKTTYPE_VIDEOSYNCHREP)
            {
                // m_lastPktVideoSynch = *reinterpret_cast<const cbPKT_VIDEOSYNCH*>(pPkt);
                if (m_pLateCallback[CBSDKCALLBACK_ALL])
                    m_pLateCallback[CBSDKCALLBACK_ALL](m_nInstance, cbSdkPkt_SYNCH, pPkt, m_pLateCallbackParams[CBSDKCALLBACK_ALL]);
                // Late bind before usage
                LateBindCallback(CBSDKCALLBACK_SYNCH);
                if (m_pLateCallback[CBSDKCALLBACK_SYNCH])
                    m_pLateCallback[CBSDKCALLBACK_SYNCH](m_nInstance, cbSdkPkt_SYNCH, pPkt, m_pLateCallbackParams[CBSDKCALLBACK_SYNCH]);
            }
            else if (pPkt->type == cbPKTTYPE_LOGREP)
            {
                if (m_pLateCallback[CBSDKCALLBACK_ALL])
                    m_pLateCallback[CBSDKCALLBACK_ALL](m_nInstance, cbSdkPkt_LOG, pPkt, m_pLateCallbackParams[CBSDKCALLBACK_ALL]);
                // Late bind before usage
                LateBindCallback(CBSDKCALLBACK_LOG);
                if (m_pLateCallback[CBSDKCALLBACK_LOG])
                    m_pLateCallback[CBSDKCALLBACK_LOG](m_nInstance, cbSdkPkt_LOG, pPkt, m_pLateCallbackParams[CBSDKCALLBACK_LOG]);
            }
        } // end if (pPkt->chid==0x8000
    } // end if (pPkt->chid & 0x8000
    else if (pPkt->chid == 0)
    {
        // No mask applied here
        // Inside the callback cbPKT_GROUP.type can be used to find the sample group number
        uint8_t smpgroup = ((cbPKT_GROUP *)pPkt)->type; // sample group
        if (smpgroup > 0 && smpgroup <= cbMAXGROUPS)
        {
            if (m_pLateCallback[CBSDKCALLBACK_ALL])
                m_pLateCallback[CBSDKCALLBACK_ALL](m_nInstance, cbSdkPkt_CONTINUOUS, pPkt, m_pLateCallbackParams[CBSDKCALLBACK_ALL]);
            // Late bind before usage
            LateBindCallback(CBSDKCALLBACK_CONTINUOUS);
            if (m_pLateCallback[CBSDKCALLBACK_CONTINUOUS])
                m_pLateCallback[CBSDKCALLBACK_CONTINUOUS](m_nInstance, cbSdkPkt_CONTINUOUS, pPkt, m_pLateCallbackParams[CBSDKCALLBACK_CONTINUOUS]);
        }
    }
    // check for channel event packets cerebus channels 1-144
    else if (pPkt->chid < cbNUM_ANALOG_CHANS + 1)   // channels are 1 based
    {
        if (pPkt->chid > 0 )
        {
            if (m_pLateCallback[CBSDKCALLBACK_ALL])
                m_pLateCallback[CBSDKCALLBACK_ALL](m_nInstance, cbSdkPkt_SPIKE, pPkt, m_pLateCallbackParams[CBSDKCALLBACK_ALL]);
            // Late bind before usage
            LateBindCallback(CBSDKCALLBACK_SPIKE);
            if (m_pLateCallback[CBSDKCALLBACK_SPIKE])
                m_pLateCallback[CBSDKCALLBACK_SPIKE](m_nInstance, cbSdkPkt_SPIKE, pPkt, m_pLateCallbackParams[CBSDKCALLBACK_SPIKE]);
        }
    }
    // catch digital input port events and save them as NSAS experiment event packets
    else if (pPkt->chid == MAX_CHANS_DIGITAL_IN)
    {
        // if (m_bChannelMask[pPkt->chid - 1])
        {
            if (m_pLateCallback[CBSDKCALLBACK_ALL])
                m_pLateCallback[CBSDKCALLBACK_ALL](m_nInstance, cbSdkPkt_DIGITAL, pPkt, m_pLateCallbackParams[CBSDKCALLBACK_ALL]);
            // Late bind before usage
            LateBindCallback(CBSDKCALLBACK_DIGITAL);
            if (m_pLateCallback[CBSDKCALLBACK_DIGITAL])
                m_pLateCallback[CBSDKCALLBACK_DIGITAL](m_nInstance, cbSdkPkt_DIGITAL, pPkt, m_pLateCallbackParams[CBSDKCALLBACK_DIGITAL]);
        }
    }
    // catch serial input port events and save them as NSAS experiment event packets
    else if (pPkt->chid == MAX_CHANS_SERIAL)
    {
        // if (m_bChannelMask[pPkt->chid - 1])
        {
            if (m_pLateCallback[CBSDKCALLBACK_ALL])
                m_pLateCallback[CBSDKCALLBACK_ALL](m_nInstance, cbSdkPkt_SERIAL, pPkt, m_pLateCallbackParams[CBSDKCALLBACK_ALL]);
            // Late bind before usage
            LateBindCallback(CBSDKCALLBACK_SERIAL);
            if (m_pLateCallback[CBSDKCALLBACK_SERIAL])
                m_pLateCallback[CBSDKCALLBACK_SERIAL](m_nInstance, cbSdkPkt_SERIAL, pPkt, m_pLateCallbackParams[CBSDKCALLBACK_SERIAL]);
        }
    }

    // save the timestamp to overcome the case where the reset button is pressed
    // (or recording started) which resets the timestamp to 0, but Central doesn't
    // reset it's timer to 0 for cbGetSystemClockTime
    m_uCbsdkTime = pPkt->time;

}
