#include "gtest/gtest.h"
#include "CereConn.h"


// The fixture for testing class Project1. From google test primer.
class CereConnBasicTest : public ::testing::Test
{
protected:
    CereConnBasicTest()
    {
        spdlog::set_pattern("[%Y-%m-%d %H:%M:%S.%f] [%n] [%^%l%$] [PID:%P] [TID:%t] %v");
    }

    cc::CereConn ck;
};


TEST_F(CereConnBasicTest, OpenCereConn)
{

    auto start = std::chrono::high_resolution_clock::now();
    auto finish = start + std::chrono::milliseconds(5000);
    
    ck.ccOpen();
    while ((std::chrono::high_resolution_clock::now() < finish) && ck.getState() != cc::CereConn::State::Idle)
    {
        std::this_thread::sleep_for(std::chrono::milliseconds(10));
    }
    ASSERT_EQ(ck.getState(), cc::CereConn::State::Idle);
}

TEST_F(CereConnBasicTest, StateTransitions)
{

    auto start = std::chrono::high_resolution_clock::now();
    auto finish = start + std::chrono::milliseconds(5000);
    ck.ccOpen();
    while ((std::chrono::high_resolution_clock::now() < finish) && ck.getState() != cc::CereConn::State::Idle)
    {
        std::this_thread::sleep_for(std::chrono::milliseconds(10));
    }
    ASSERT_EQ(ck.getState(), cc::CereConn::State::Idle);
    
    start = std::chrono::high_resolution_clock::now();
    finish = start + std::chrono::milliseconds(5000);
    
    ck.ccClose();
    
    while ((std::chrono::high_resolution_clock::now() < finish) && ck.getState() != cc::CereConn::State::Closed)
    {
        std::this_thread::sleep_for(std::chrono::milliseconds(10));
    }
    
    ASSERT_EQ(ck.getState(), cc::CereConn::State::Closed);
    
    start = std::chrono::high_resolution_clock::now();
    finish = start + std::chrono::milliseconds(5000);
    
    ck.ccOpen();
    while ((std::chrono::high_resolution_clock::now() < finish) && ck.getState() != cc::CereConn::State::Idle)
    {
        std::this_thread::sleep_for(std::chrono::milliseconds(10));
    }
    ASSERT_EQ(ck.getState(), cc::CereConn::State::Idle);
    
    start = std::chrono::high_resolution_clock::now();
    finish = start + std::chrono::milliseconds(5000);
    
    ck.ccRecord();
    while ((std::chrono::high_resolution_clock::now() < finish) && ck.getState() != cc::CereConn::State::Recording)
    {
        std::this_thread::sleep_for(std::chrono::milliseconds(10));
    }
    ASSERT_EQ(ck.getState(), cc::CereConn::State::Recording);
    
    std::this_thread::sleep_for(std::chrono::milliseconds(500));
    
    start = std::chrono::high_resolution_clock::now();
    finish = start + std::chrono::milliseconds(5000);
    
    ck.ccIdle();
    while ((std::chrono::high_resolution_clock::now() < finish) && ck.getState() != cc::CereConn::State::Idle)
    {
        std::this_thread::sleep_for(std::chrono::milliseconds(10));
    }
    ASSERT_EQ(ck.getState(), cc::CereConn::State::Idle);
    
}



TEST_F(CereConnBasicTest, CloseCereConn)
{
    ck.ccOpen();
    auto start = std::chrono::high_resolution_clock::now();
    auto finish = start + std::chrono::milliseconds(5000);
    
    while ((std::chrono::high_resolution_clock::now() < finish) && ck.getState() != cc::CereConn::State::Idle)
    {
        std::this_thread::sleep_for(std::chrono::milliseconds(10));
    }
    ck.ccClose();

    start = std::chrono::high_resolution_clock::now();
    finish = start + std::chrono::milliseconds(5000);
    while ((std::chrono::high_resolution_clock::now() < finish) && ck.getState() != cc::CereConn::State::Closed)
    {
        std::this_thread::sleep_for(std::chrono::milliseconds(10));
    }
    ASSERT_EQ(ck.getState(), cc::CereConn::State::Closed);
}


TEST_F(CereConnBasicTest, SpikeRateEstimator)
{
    ck.ccOpen();
    ck.fillSpikeRateEstimatorChannelUnitList(5, 3);
    {
        auto ch_u_m_pair = ck.getSpikeRateEstimatorChannelUnitMapping();
    
        auto ch_u_list = ch_u_m_pair.first;
        auto ch_u_map = ch_u_m_pair.second;
    
        EXPECT_EQ(ch_u_list.size(), 15);
        EXPECT_EQ(ch_u_map.size(), 15);
    
        EXPECT_EQ(ch_u_list[0].first, 1);
        EXPECT_EQ(ch_u_list[0].second, 0);
        EXPECT_EQ(ch_u_list[1].first, 1);
        EXPECT_EQ(ch_u_list[1].second, 1);
        EXPECT_EQ(ch_u_list[2].first, 1);
        EXPECT_EQ(ch_u_list[2].second, 2);
        EXPECT_EQ(ch_u_list[3].first, 2);
        EXPECT_EQ(ch_u_list[3].second, 0);
    
        EXPECT_EQ((ch_u_map.at(std::make_pair(1, 0))), 0);
        EXPECT_EQ((ch_u_map.at(std::make_pair(2, 1))), 4);
        EXPECT_EQ((ch_u_map.at(std::make_pair(3, 0))), 6);
        EXPECT_EQ((ch_u_map.at(std::make_pair(4, 2))), 11);
        EXPECT_THROW((ch_u_map.at(std::make_pair(6, 2))), std::exception);
    }
    {
        auto ch_u_list = ck.getSpikeRateEstimatorChannelUnitList();
        EXPECT_EQ(ch_u_list[0].first, 1);
        EXPECT_EQ(ch_u_list[0].second, 0);
        EXPECT_EQ(ch_u_list[1].first, 1);
        EXPECT_EQ(ch_u_list[1].second, 1);
        EXPECT_EQ(ch_u_list[2].first, 1);
        EXPECT_EQ(ch_u_list[2].second, 2);
        EXPECT_EQ(ch_u_list[3].first, 2);
        EXPECT_EQ(ch_u_list[3].second, 0);
        
    }

    ck.setSpikeRateEstimatorChannelUnitList({{6, 3}, {1,0}, {2,3}, {1,3}});
    {
        auto ch_u_m_pair = ck.getSpikeRateEstimatorChannelUnitMapping();
    
        auto ch_u_list = ch_u_m_pair.first;
        auto ch_u_map = ch_u_m_pair.second;
    
        EXPECT_EQ(ch_u_list.size(), 4);
        EXPECT_EQ(ch_u_map.size(), 4);
    
        EXPECT_EQ(ch_u_list[0].first, 6);
        EXPECT_EQ(ch_u_list[0].second, 3);
        EXPECT_EQ(ch_u_list[1].first, 1);
        EXPECT_EQ(ch_u_list[1].second, 0);
        EXPECT_EQ(ch_u_list[2].first, 2);
        EXPECT_EQ(ch_u_list[2].second, 3);
        EXPECT_EQ(ch_u_list[3].first, 1);
        EXPECT_EQ(ch_u_list[3].second, 3);
    
        EXPECT_EQ((ch_u_map.at(std::make_pair(6, 3))), 0);
        EXPECT_EQ((ch_u_map.at(std::make_pair(1, 0))), 1);
        EXPECT_EQ((ch_u_map.at(std::make_pair(2, 3))), 2);
        EXPECT_EQ((ch_u_map.at(std::make_pair(1, 3))), 3);
        EXPECT_THROW((ch_u_map.at(std::make_pair(6, 2))), std::exception);
    }
    {
        auto ch_u_list = ck.getSpikeRateEstimatorChannelUnitList();
        EXPECT_EQ(ch_u_list[0].first, 6);
        EXPECT_EQ(ch_u_list[0].second, 3);
        EXPECT_EQ(ch_u_list[1].first, 1);
        EXPECT_EQ(ch_u_list[1].second, 0);
        EXPECT_EQ(ch_u_list[2].first, 2);
        EXPECT_EQ(ch_u_list[2].second, 3);
        EXPECT_EQ(ch_u_list[3].first, 1);
        EXPECT_EQ(ch_u_list[3].second, 3);
    }
    
}

