#include "gtest/gtest.h"

#include <array>
#include <vector>
#include "CereConnHelpers.h"


TEST(CereConnHelpersTest, IsEqual)
{
    std::vector<int> v1{1,2,4};
    std::vector<int> v1a{1,2,4};
    std::vector<unsigned int> v2{1,2,4};
    std::vector<double> v3{1,2,4};
    std::vector<int> v4{1,2,4,0};
    std::vector<int> v5{1,2,4,0};
    std::vector<int> w1{1,2};
    std::vector<int> w2{1,2,4,5};
    std::vector<int> w3{1,3,4};
    
    
    std::array<int, 5> z1{1,2,4,0,0};
    std::array<int, 5> z2{1,2,0,0,0};
    std::array<int, 5> z3{1,2,4,3,0};
    
    EXPECT_TRUE(cc::it_chain_equal(v1, v1));
    EXPECT_TRUE(cc::it_chain_equal(v1, v1a));
    EXPECT_TRUE(cc::it_chain_equal(v1, v2));
    EXPECT_TRUE(cc::it_chain_equal(v1, v3));
    EXPECT_TRUE(cc::it_chain_equal(v1, v4));
    EXPECT_TRUE(cc::it_chain_equal(v1, z1));
    EXPECT_TRUE(cc::it_chain_equal(v4, z1));
    EXPECT_TRUE(cc::it_chain_equal(v4, v5));
    EXPECT_FALSE(cc::it_chain_equal(v1, w1));
    EXPECT_FALSE(cc::it_chain_equal(v1, w2));
    EXPECT_FALSE(cc::it_chain_equal(v1, w3));
    EXPECT_FALSE(cc::it_chain_equal(v1, z2));
    EXPECT_FALSE(cc::it_chain_equal(v1, z3));
}

TEST(CereConnHelpersTest, ItRangesEq)
{

    std::vector<int> v1{1,2,4};
    std::vector<int> v1a{1,2,4};
    std::vector<unsigned int> v2{1,2,4};
    std::vector<double> v3{1,2,4};
    std::vector<int> v4{1,2,4,0};
    std::vector<int> v5{1,2,4,0};
    std::vector<int> w1{1,2};
    std::vector<int> w2{1,2,4,5};
    std::vector<int> w3{1,3,4};


    std::array<int, 5> z1{1,2,4,0,0};
    std::array<int, 5> z2{1,2,0,0,0};
    std::array<int, 5> z3{1,2,4,3,0};

    EXPECT_TRUE(cc::it_chain_equal(v1.cbegin(),v1.cend(), v1.cbegin(), v1.cend()));
    EXPECT_TRUE(cc::it_chain_equal(v1.cbegin(),v1.cend(), v1a.cbegin(), v1a.cend()));
    EXPECT_TRUE(cc::it_chain_equal(v1.cbegin(),v1.cend(), v2.cbegin(), v2.cend()));
    EXPECT_TRUE(cc::it_chain_equal(v1.cbegin(),v1.cend(), v3.cbegin(), v3.cend()));
    EXPECT_TRUE(cc::it_chain_equal(v1.cbegin(),v1.cend(), v4.cbegin(), v4.cend()));
    EXPECT_TRUE(cc::it_chain_equal(v1.cbegin(),v1.cend(), z1.cbegin(), z1.cend()));
    EXPECT_TRUE(cc::it_chain_equal(v4.cbegin(),v4.cend(), z1.cbegin(), z1.cend()));
    EXPECT_TRUE(cc::it_chain_equal(v4.cbegin(),v4.cend(), v5.cbegin(), v5.cend()));
    EXPECT_FALSE(cc::it_chain_equal(v1.cbegin(), v1.cend(), w1.cbegin(), w1.cend()));
    EXPECT_FALSE(cc::it_chain_equal(v1.cbegin(), v1.cend(), w2.cbegin(), w2.cend()));
    EXPECT_FALSE(cc::it_chain_equal(v1.cbegin(), v1.cend(), w3.cbegin(), w3.cend()));
    EXPECT_FALSE(cc::it_chain_equal(v1.cbegin(), v1.cend(), z2.cbegin(), z2.cend()));
    EXPECT_FALSE(cc::it_chain_equal(v1.cbegin(), v1.cend(), z3.cbegin(), z3.cend()));

}