#include "gtest/gtest.h"
#include "CereConn.h"


namespace cc
{
    
TEST(CereConnSBPE, SimpleInit)
{
    spdlog::set_level(spdlog::level::info);
    spdlog::set_pattern("[%Y-%m-%d %H:%M:%S.%f] [%n] [%^%l%$] [PID:%P] [TID:%t] %v");
    
    CereConn ck{0, false, true};
    
    ck.ccOpen();
    std::this_thread::sleep_for(std::chrono::milliseconds(50));
    ck.ccRecord();
    std::this_thread::sleep_for(std::chrono::milliseconds(100));
    
    sbpe::SpikeBandPowerEstimator::sbp_list_type sbpl;
    uint32_t cbTime;
    ck.getSpikeBandPowerData(sbpl, cbTime);
    std::this_thread::sleep_for(std::chrono::milliseconds(200));
    ck.getSpikeBandPowerData(sbpl, cbTime);
    
    ASSERT_GE(sbpl.size(), 1);
    ASSERT_GE(sbpl[0].second.size(), 1);
    EXPECT_GE(sbpl[0].second[0], 0.1);
    
}

TEST(CereConnSBPE, ExplicitInit)
{
    spdlog::set_level(spdlog::level::info);
    spdlog::set_pattern("[%Y-%m-%d %H:%M:%S.%f] [%n] [%^%l%$] [PID:%P] [TID:%t] %v");
    
    CereConn ck{0, false, false};
    ck.spikeBandPowerEstimatorStart();
    ck.ccOpen();
    std::this_thread::sleep_for(std::chrono::milliseconds(50));
    ck.ccRecord();
    std::this_thread::sleep_for(std::chrono::milliseconds(100));
    
    sbpe::SpikeBandPowerEstimator::sbp_list_type sbpl;
    uint32_t cbTime;
    ck.getSpikeBandPowerData(sbpl, cbTime);
    std::this_thread::sleep_for(std::chrono::milliseconds(200));
    ck.getSpikeBandPowerData(sbpl, cbTime);
    
    ASSERT_GE(sbpl.size(), 1);
    ASSERT_GE(sbpl[0].second.size(), 1);
    EXPECT_GE(sbpl[0].second[0], 0.1);
    
}


TEST(CereConnSBPE, SetSBPEFilterSamples)
{
    spdlog::set_level(spdlog::level::info);
    spdlog::set_pattern("[%Y-%m-%d %H:%M:%S.%f] [%n] [%^%l%$] [PID:%P] [TID:%t] %v");
    
    CereConn ck{0, false, true};
    // set RMS sample period to 10ms
    ck.setSpikeBandPowerEstimatorIIRFilterNSamples(300);
    ASSERT_EQ(ck.getSpikeBandPowerEstimatorIIRFilterNSamples(), 300);
    ck.ccOpen();
    std::this_thread::sleep_for(std::chrono::milliseconds(50));
    ck.ccRecord();
    std::this_thread::sleep_for(std::chrono::milliseconds(100));
    
    sbpe::SpikeBandPowerEstimator::sbp_list_type sbpl;
    uint32_t cbTime;
    ck.getSpikeBandPowerData(sbpl, cbTime);
    std::this_thread::sleep_for(std::chrono::milliseconds(200));
    ck.getSpikeBandPowerData(sbpl, cbTime);
    
    ASSERT_GE(sbpl.size(), 19);
    ASSERT_GE(sbpl[0].second.size(), 1);
    EXPECT_GE(sbpl[0].second[0], 0.1);
    
}

TEST(CereConnSBPE, SBPEOverflow)
{
    spdlog::set_level(spdlog::level::info);
    spdlog::set_pattern("[%Y-%m-%d %H:%M:%S.%f] [%n] [%^%l%$] [PID:%P] [TID:%t] %v");

    CereConn ck{0, false, true};
    // set RMS sample period to 10ms
    ck.setSpikeBandPowerEstimatorIIRFilterNSamples(30);
    ASSERT_EQ(ck.getSpikeBandPowerEstimatorIIRFilterNSamples(), 30);
    ck.setSpikeBandPowerEstimatorLoopIntervalMS(1);
    ck.ccOpen();
    std::this_thread::sleep_for(std::chrono::milliseconds(50));
    ck.ccRecord();
    std::this_thread::sleep_for(std::chrono::milliseconds(100));

    sbpe::SpikeBandPowerEstimator::sbp_list_type sbpl;
    uint32_t cbTime;
    ck.getSpikeBandPowerData(sbpl, cbTime);
    std::this_thread::sleep_for(std::chrono::milliseconds(4000));
    ck.getSpikeBandPowerData(sbpl, cbTime);

    ASSERT_GE(sbpl.size(), 1000);
    std::this_thread::sleep_for(std::chrono::milliseconds(4000));
    ck.getSpikeBandPowerData(sbpl, cbTime);
    ASSERT_GE(sbpl.size(), 1000);

}



} /* cc */