#include <cstdio>
#include "gtest/gtest.h"
#include "CereRingBuffer.h"

template<size_t Size>
using TestCommentRingBuffer = cc::CommentRingBufferT<cc::Timestamp_t, cc::constants::CommentLength, Size>;

using TestCommentList = cc::CommentListT<cc::Timestamp_t, cc::constants::CommentLength>;



// The fixture for testing class Project1. From google test primer.
class CCRBTest : public ::testing::Test
{
protected:
    // Objects declared here can be used by all tests in the test case for Project1.
    static constexpr size_t TRBSize = 5;
    
    TestCommentRingBuffer<TRBSize> ccrb;
    
    cbPKT_COMMENT pc01 = {17, 0x8000, cbPKTTYPE_COMMENTREP, cbPKTDLEN_COMMENT, {0, cbCOMMENT_FLAG_RGBA, {0, 0}}, 0xff2244, "A comment"};
    cbPKT_COMMENT pc02 = {117, 0x8000, cbPKTTYPE_COMMENTREP, cbPKTDLEN_COMMENT, {0, cbCOMMENT_FLAG_RGBA, {0, 0}}, 0x442244, "Another comment"};
    
};
constexpr size_t CCRBTest::TRBSize;


TEST_F(CCRBTest, RingBufferEmpty)
{
    ASSERT_TRUE(ccrb.wasEmpty());
}

TEST_F(CCRBTest, AddOneComment)
{
    ccrb.addComment(pc01);
    ASSERT_FALSE(ccrb.wasEmpty());
}

TEST_F(CCRBTest, ReadOneComment)
{
    TestCommentList cbuf;
    EXPECT_EQ(cbuf.size(), 0);
    
    ccrb.addComment(pc01);
    ccrb.consume(cbuf);
    EXPECT_EQ(cbuf.size(), 1);
    
    ASSERT_EQ(cbuf[0].timestamp, 17);
    ASSERT_EQ(cbuf[0].data, 0xff2244);
    ASSERT_EQ(cbuf[0].flag, cbCOMMENT_FLAG_RGBA);
    ASSERT_STREQ(cbuf[0].comment, "A comment");
    
    ASSERT_TRUE(ccrb.wasEmpty());
}



TEST_F(CCRBTest, ReadTwoComments)
{
    int read_start, read_end, num_samples;
    TestCommentList cbuf;
    
    ccrb.addComment(pc01);
    ccrb.addComment(pc02);
    ccrb.consume(cbuf);
    EXPECT_EQ(cbuf.size(), 2);
    
    ASSERT_EQ(cbuf[0].timestamp, 17);
    ASSERT_EQ(cbuf[0].data, 0xff2244);
    ASSERT_EQ(cbuf[0].flag, cbCOMMENT_FLAG_RGBA);
    ASSERT_STREQ(cbuf[0].comment, "A comment");
    
    ASSERT_EQ(cbuf[1].timestamp, 117);
    ASSERT_EQ(cbuf[1].data, 0x442244);
    ASSERT_EQ(cbuf[1].flag, cbCOMMENT_FLAG_RGBA);
    ASSERT_STREQ(cbuf[1].comment, "Another comment");
    
    ASSERT_TRUE(ccrb.wasEmpty());
    
}

TEST_F(CCRBTest, AddTooManyComments)
{
    for (size_t i = 1; i <= TRBSize; ++i)
    {
        pc01.time = i;
        sprintf(pc01.comment, "Comment #%06lu", i);
        ccrb.addComment(pc01);
    }
    ASSERT_EQ(ccrb.getOverflowCount(), 0);
    
    TestCommentList cbuf;
    
    ++ pc01.time;
    sprintf(pc01.comment, "Comment #%06u", pc01.time);
    ASSERT_FALSE(ccrb.addComment(pc01));
    ASSERT_EQ(ccrb.getOverflowCount(), 1);

    ccrb.consume(cbuf);
    EXPECT_EQ(cbuf.size(), TRBSize);
    
    ASSERT_EQ(cbuf[0].timestamp, 1);
    ASSERT_STREQ(cbuf[0].comment, "Comment #000001");
    ASSERT_EQ(cbuf[TRBSize - 1].timestamp, TRBSize);
    
}
