#include <cstdio>
#include "gtest/gtest.h"
#include "CereRingBuffer.h"

template<size_t Size>
using TestDigInEvRingBuffer = cc::DigInEvRingBufferT<cc::Timestamp_t, Size>;

using TestDigInEvList = cc::DigInEvListT<cc::Timestamp_t>;

// The fixture for testing class Project1. From google test primer.
class DIERBTest : public ::testing::Test
{
protected:
    // Objects declared here can be used by all tests in the test case for Project1.
    static constexpr size_t TRBSize = 5;
    
    TestDigInEvRingBuffer<TRBSize> dierb;
    
    cbPKT_DINP pdie01 = {17, cc::constants::ChannelDigitalIn, 0, 1, 0x0000001f};
    cbPKT_DINP pdie02 = {117, cc::constants::ChannelSerialIn, 0, 1, 0x00000042};
    
};
constexpr size_t DIERBTest::TRBSize;


TEST_F(DIERBTest, RingBufferEmpty)
{
    ASSERT_TRUE(dierb.wasEmpty());
}

TEST_F(DIERBTest, AddOneDigInEv)
{
    dierb.addDigIn(pdie01);
    ASSERT_FALSE(dierb.wasEmpty());
}

TEST_F(DIERBTest, ReadOneDigInEv)
{
    TestDigInEvList diebuf;
    EXPECT_EQ(diebuf.size(), 0);
    
    dierb.addDigIn(pdie01);
    dierb.consume(diebuf);
    EXPECT_EQ(diebuf.size(), 1);
    
    ASSERT_EQ(diebuf[0].timestamp, 17);
    ASSERT_EQ(diebuf[0].data, 0x1f);
    ASSERT_EQ(diebuf[0].source, cc::DigInEv::Source::DigIn);
    
    ASSERT_TRUE(dierb.wasEmpty());
}

TEST_F(DIERBTest, ReadTwoDigInEvents)
{
    int read_start, read_end, num_samples;
    TestDigInEvList diebuf;
    
    dierb.addDigIn(pdie01);
    dierb.addDigIn(pdie02);
    dierb.consume(diebuf);
    EXPECT_EQ(diebuf.size(), 2);
    
    ASSERT_EQ(diebuf[0].timestamp, 17);
    ASSERT_EQ(diebuf[0].data, 0x1f);
    ASSERT_EQ(diebuf[0].source, cc::DigInEv::Source::DigIn);
    
    ASSERT_EQ(diebuf[1].timestamp, 117);
    ASSERT_EQ(diebuf[1].data, 0x42);
    ASSERT_EQ(diebuf[1].source, cc::DigInEv::Source::Serial);
    
    ASSERT_TRUE(dierb.wasEmpty());
    
}

TEST_F(DIERBTest, AddTooManyDigInEvents)
{
    for (size_t i = 1; i <= TRBSize; ++i)
    {
        pdie01.time = i;
        pdie01.data[0] = i;
        dierb.addDigIn(pdie01);
    }
    ASSERT_EQ(dierb.getOverflowCount(), 0);
    
    TestDigInEvList diebuf;
    
    ++ pdie01.time;
    pdie01.data[0] = pdie01.time;
    ASSERT_FALSE(dierb.addDigIn(pdie01));
    ASSERT_EQ(dierb.getOverflowCount(), 1);

    dierb.consume(diebuf);
    EXPECT_EQ(diebuf.size(), TRBSize);
    
    ASSERT_EQ(diebuf[0].timestamp, 1);
    ASSERT_EQ(diebuf[0].data, 1);
    ASSERT_EQ(diebuf[TRBSize - 1].timestamp, TRBSize);
    
}
