from sklearn.discriminant_analysis import LinearDiscriminantAnalysis as LDA
from sklearn.metrics import classification_report
from sklearn.model_selection import GridSearchCV

scores = ['precision', 'recall']

# clf1 = LDA(solver=solver, shrinkage=shrinkage)

tuned_parameters = [{'shrinkage': list(np.arange(0.1,0.91,0.1)), 'solver': ['lsqr', 'eigen']}]

data_train = clf.train_trials
n_features = data_train[0, 0].shape[0]
n_classes = 3
X_tot = np.empty((0, n_features))
y_tot = np.empty((0, 1))


for ii in range(n_classes):
    X_tot = np.vstack((X_tot, data_train[ii, 0].T))
    y_tot = np.vstack((y_tot, data_train[ii, 0][0][:, None] * 0 + ii))

X_train = X_tot
y_train = y_tot

X_test = X_tot
y_test = y_tot


kf = StratifiedShuffleSplit(n_splits=10, test_size=0.2, random_state=None)

for score in scores:
    print("# Tuning hyper-parameters for %s" % score)
    print()

    clf3 = GridSearchCV(LDA(), tuned_parameters, cv=kf, n_jobs=4,
                       scoring='%s_macro' % score)
    clf3.fit(X_train, y_train.ravel())
    print(f'Best params: {clf3.best_params_}\n')
    means = clf3.cv_results_['mean_test_score']
    stds = clf3.cv_results_['std_test_score']

    for mean, std, params in zip(means, stds, clf3.cv_results_['params']):
        print("%0.3f (+/-%0.03f) for %r"
              % (mean, std * 2, params))

    y_pred = clf3.predict(X_test)
    print(classification_report(y_test, y_pred))
    print()
