'''
description: parameter scan
author: Ioannis Vlachos
date:   13.12.18

Copyright (c) 2018  Ioannis Vlachos
All rights reserved.'''


import importlib
from datetime import datetime as dt

# import copy
import matplotlib.pyplot as plt
# import pickle
import numpy as np
# import sys
import scipy.signal as signal
from scipy import io
from sklearn.metrics import auc, roc_curve
from sklearn.model_selection import ParameterGrid

import analytics
import aux
import modules.classifier as clf
from aux import log
from helpers import data_management as dm

importlib.reload(aux)
importlib.reload(clf)
importlib.reload(dm)
importlib.reload(analytics)



aux.config_logging()

# param_grid = {'template': [[10,30,50]], 'reg_fact': np.arange(0.0, 1.01, .2)}
# param_grid = {'template': [np.arange(10,50).tolist()], 'reg_fact': np.arange(0.1, 0.11, .2)}
param_grid = {'template': [np.arange(10,20).tolist()], 'reg_fact': np.arange(0.1, 0.11, .2)}
par_list = list(ParameterGrid(param_grid))


params = aux.load_config()
params.classifier.online = False
save_model = params.classifier.model_training

solver = params.classifier.model_training.solver

# features_fname = '/kiap/data/tom/train_trials.pkl'
features_fname = params.file_handling.datafile_path + '/train_trials.pkl'    # name of features array
model_fname01 = params.file_handling.datafile_path + '/model1'               # save sklearn LDA model
model_fname02 = params.file_handling.datafile_path + '/model2'               # save explicit LDA model


data_tot, tt, triggers = dm.get_raw(n_triggers=params.classifier.n_classes)
data_tot = analytics.exclude_channels(data_tot, params.classifier.channel_mask)         # exclude channels from analysis

session_ids = list(range(0, data_tot.shape[0]))         # list with sessions
clf = clf.Classifier(params)

# triggers[0,1] = triggers[0,1][0:1, 1:-1]          # remove some triggers if necessary, use [0:1, ...] to have consistent format
# triggers[0,2] = triggers[0,2][0:1, 1:-1]  

clf.triggers_shape = triggers.shape
clf.triggers = triggers


fct = params.daq.spike_rates.loop_interval / 1000.      # convert samples to sec

# PEAK FINDER
pf_threshold = params.classifier.peaks.height                # threshold of probability
pf_width = params.classifier.peaks.width                     # consecutive bins above threshold
pf_distance = params.recording.timing.t_response/fct         # distance between consecutive peaks


# PLOTTING PARAMETERS
# triggers_plot = triggers.shape[1]
triggers_plot = 3




tp_tot = np.zeros((len(par_list), params.classifier.n_classes))       # tp, fp, tn
fp_tot = np.zeros((len(par_list), params.classifier.n_classes))       # tp, fp, tn
tn_tot = np.zeros((len(par_list), params.classifier.n_classes))       # tp, fp, tn

tpr_tot = np.zeros((len(par_list), params.classifier.n_classes))       # tp, fp, tn
fpr_tot = np.zeros((len(par_list), params.classifier.n_classes))       # tp, fp, tn
tnr_tot = np.zeros((len(par_list), params.classifier.n_classes))       # tp, fp, tn

peaks_tot = []

cnt = 0
for par_id, pars in enumerate(par_list):

    params.classifier.template = np.array(pars['template'])
    params.classifier.reg_fact = pars['reg_fact']


    session_id = 0
    sess = 0
    cur_data = data_tot[session_id:session_id + 1]
    # cur_data = data_tot[-2:-1]          # get last session for testing

    tp = np.zeros((len(par_list), params.classifier.n_classes))
    fp = np.zeros((len(par_list), params.classifier.n_classes))
    tn = np.zeros((len(par_list), params.classifier.n_classes))

    model_fname1 = model_fname01 + '_' + dt.now().strftime('%H%M%S') + '.pkl'
    model_fname2 = model_fname02 + '_' + dt.now().strftime('%H%M%S') + '.pkl'

    clf.get_trials_kiap(data_tot[session_ids, :], triggers, features_fname=features_fname)                      # 1. GET TRAIN FEATURES
    _, _, _, _ = clf.train_LDA(clf.train_trials, baseline=0, features_fname=features_fname,                     # 2. TRAIN MODEL
                        # model_fname1=model_fname1, model_fname2=model_fname2, solver='lsqr', save_model=save_model)
                        model_fname1=model_fname1, model_fname2=model_fname2, solver=solver, save_model=save_model)
    test_trials = clf.get_test_trials_kiap(cur_data)                                                            # 3. GET TEST FEATURES
    clf.test_LDA_kiap(test_trials, session_id=sess, compare_results=0)                                          # 4. TEST MODEL

    xx1 = np.arange(cur_data[0, 0].shape[0]) * fct
    xx2 = np.arange(test_trials.shape[0]) * fct


    if 'explicit' in params.classifier.model_training.model:
        prob = clf.rel_prob     # explicit LDA
    elif 'scikit' in params.classifier.model_training.model:
        prob = clf.prob         # sklearn model

    if prob.shape[1] == 2:      # to be sure that beginning and end are also detected
        prob[0, 1] = 0
        prob[-1, 1] = 0

    # CALCULATE CLASSIFICATION PERFORMANCE
    for cl_id in range(triggers_plot):
        peaks, _ = signal.find_peaks(prob[:, cl_id], height=pf_threshold, width=pf_width, distance=pf_distance)
        for idx in peaks:
            if np.any(np.abs(xx2[idx] - triggers[0, cl_id] * fct) < params.recording.timing.t_response+1):
                tp[par_id, cl_id] += 1
            else:
                fp[par_id, cl_id] += 1

        tn[par_id, cl_id] = triggers[0, cl_id].size - tp[par_id, cl_id]

        peaks_tot.append(peaks)

        tpr_tot[par_id, cl_id] = tp[par_id, cl_id] / triggers[0, cl_id].size
        fpr_tot[par_id, cl_id] = fp[par_id, cl_id] / (fp[par_id, cl_id] + sum([triggers[0, ii].size for ii in range(params.classifier.n_classes)]) - triggers[0, cl_id].size)
        # fnr_tot[par_id, cl_id] = fn[par_id, cl_id] / (fn[par_id, cl_id] + tp[par_id, cl_id])
        tnr_tot[par_id, cl_id] = 1 - fpr_tot[par_id, cl_id]


    tp_tot[cnt, :] = tp[par_id]
    fp_tot[cnt, :] = fp[par_id]
    tn_tot[cnt, :] = tn[par_id]
    cnt += 1

for ii in range(cnt):
    # np.set_printoptions(precision=2)
    np.set_printoptions(precision=2, formatter={'int': lambda x: '%2d' % x, 'float': lambda x: '%.2f' % x}, )
    print(par_list[ii])
    print('tp', tp_tot[ii,:], tpr_tot[ii,:])
    print('fp', fp_tot[ii,:], fpr_tot[ii,:])
    print('tn', tn_tot[ii,:], tnr_tot[ii,:])
    print()
    np.set_printoptions(precision=8)

a = np.argwhere(tp_tot[:, 0] == np.amax(tp_tot[:, 0]))
b = np.argwhere(fp_tot[:, 0] == np.amin(fp_tot[:, 0]))
best_pars = [par_list[idx] for idx in np.intersect1d(a, b)]
print(best_pars)



# PLOT RESULTS
# ------------

plt.figure(1)
plt.clf()
ax1 = plt.subplot(311)
ax2 = plt.subplot(312, sharex=ax1)
ax3 = plt.subplot(313, sharex=ax1)
plt.ylim(0,2)

plt.draw()

stem_height = np.max(cur_data[0,0].flatten())*1.2

col = ['C0', 'C1', 'C2']
for cl_id in range(triggers_plot):
    markerline, stemlines, baseline = ax1.stem(triggers[0, cl_id][0] * fct, triggers[0, cl_id][0] * 0 + stem_height, '-', basefmt=' ')
    plt.setp(stemlines, alpha=0.8, color=col[cl_id], lw=2)
    # plt.setp(markerline, marker=None)
    plt.setp(markerline, alpha=1, color=col[cl_id], lw=1)

    if peaks_tot[cl_id].size>0:
        markerline, stemlines, baseline = ax2.stem(xx2[peaks_tot[cl_id]], xx2[peaks_tot[cl_id]] * 0 + 2)
        plt.setp(stemlines, alpha=1, color=col[cl_id], lw=1)
        plt.setp(markerline, alpha=1, color=col[cl_id], lw=1)

    ax2.plot(xx2, clf.prob[:,cl_id], '-', lw=2, color=col[cl_id])
    ax3.plot(xx2, clf.rel_prob[:,cl_id], '-', lw=2, color=col[cl_id])

plt.gca().set_prop_cycle(color=['#d62728', '#9467bd'])
ax1.plot(xx1, cur_data[0, 0][:, :2],'-.')

# analytics.plot_psth(clf)

plt.draw()
plt.show()
