'''
description: parameter scan
author: Ioannis Vlachos
date:   13.12.18

Copyright (c) 2018  Ioannis Vlachos
All rights reserved.'''

import sys
import importlib
from datetime import datetime as dt
import os

import matplotlib.pyplot as plt
import numpy as np
import scipy.signal as signal
from scipy import io
from sklearn.metrics import auc, roc_curve
from sklearn.model_selection import ParameterGrid
from sklearn.decomposition.pca import PCA
from sklearn.externals import joblib

from analytics import analytics1
import aux
import modules.classifier as clf
from aux import log
from helpers import data_management as dm

import yaml

importlib.reload(aux)
importlib.reload(clf)
importlib.reload(dm)
importlib.reload(analytics1)

aux.config_logging()


template_step = 2
template_offset = np.arange(1, 21, 4)
template_end = np.arange(1, 40, 4)
template = [list(range(ii, ii + jj, template_step)) for ii in template_offset for jj in template_end]
# template = [np.arange(5, 35, 2)]
# template = [np.arange(20,40,2).tolist()]
# param_grid = {'template': [np.arange(1,60,4).tolist()], 'reg_fact': np.arange(0.3, 0.31, .2), 'pf_width': [[5],[10],[20],[40],[50],[60]], 'pf_height': [[70],[0.8]]}
# param_grid = {'template': [np.arange(20,60,5).tolist(), np.arange(10,50,5).tolist(),np.arange(1,80,5).tolist()], 'reg_fact': np.arange(0.1, 0.81, .2), 'pf_width': [[10],[20],[30]], 'pf_height': [[0.8]]}
# param_grid = {'template': template, 'reg_fact': np.arange(0.5, 0.51, .1), 'pf_width': [[70]], 'pf_height': [[0.8]]}
param_grid = {'template': [np.arange(1, 2, 2).tolist()], 'reg_fact': np.arange(0.1, 0.11, .1), 'pf_width': [[70]], 'pf_height': [[0.8]]}
# param_grid = {'template': [[20, 22, 24, 26, 28, 30, 32, 34, 36, 38]], 'reg_fact': np.arange(0.5, 0.51, .2), 'pf_width': [[27]], 'pf_height': [[0.8]]}
# param_grid = {'template': [[1,20,40]], 'reg_fact': np.arange(0.5, 0.51, .2)}
par_list = list(ParameterGrid(param_grid))

# par_list = [par_list[8]]

params = aux.load_config(force_reload=True)
params.classifier.online = False
save_model = params.classifier.model_training.save_model

solver = params.classifier.model_training.solver

# features_fname = '/kiap/data/tom/train_trials.pkl'
features_fname = params.file_handling.datafile_path + '/train_trials.pkl'    # name of features array
model_fname01 = params.file_handling.datafile_path + '/model1'               # save sklearn LDA model
model_fname02 = params.file_handling.datafile_path + '/model2'               # save explicit LDA model

fct = params.daq.spike_rates.loop_interval / 1000.      # convert samples to sec
idx1 = int(1/fct*params.recording.timing.t_baseline_1)      # get idx where baseline ends


# IMPORT DATA
# --------------
# data_tot: individual sessions with shape: (time, channels)

if params.speller.type == 'exploration':
    data_tot, tt, triggers_tot, ch_rec_list, file_names = dm.get_raw(n_triggers=params.classifier.n_classes,
                                                                     exploration=True, trigger_pos=params.classifier.
                                                                     trigger_pos)
elif params.speller.type == 'feedback':
    data_tot, tt, triggers_tot, ch_rec_list, file_names = dm.get_raw(n_triggers=params.classifier.n_classes,
                                                                     feedback=True, trigger_pos=params.classifier.
                                                                     trigger_pos)
else:
    data_tot, tt, triggers_tot, ch_rec_list, file_names = dm.get_raw(n_triggers=params.classifier.n_classes,
                                                                     trigger_pos=params.classifier.trigger_pos)

if params.classifier.max_active_ch_nr != []:
    params.classifier.include_channels = np.argsort(np.max(data_tot[0,0],axis=0))[-params.classifier.max_active_ch_nr:]      # get channels with high firing rates
if params.classifier.exclude_channels == []:
    exclude_channels = list(set(range(128))-set(params.classifier.include_channels))
else:
    exclude_channels = params.classifier.exclude_channels
data_tot = analytics1.exclude_channels(data_tot, exclude_channels)         # exclude channels from analysis
if params.daq.spike_rates.correct_bl_model:
    data_tot = analytics1.correct_init_baseline(data_tot, idx1, params.daq.spike_rates.bl_offset)  #correct baseline as well

try:
    clf = clf.Classifier(params)
except Exception as e:
    # log.warning(e)
    sys.exit()

# xx
min_rate = 2
exclude_ch = np.argwhere(data_tot[0, 0].mean(0) < min_rate)
for ii in range(len(data_tot)):
    data_tot[ii, 0] = np.delete(data_tot[ii, 0], exclude_ch, axis=1)

ch_ids2 = np.delete(np.arange(128), exclude_ch, axis=0)

# xx
# ch_ids2 = np.delete(np.arange(128), exclude_ids, axis=0)



# triggers[0,1] = triggers[0,1][0:1, 1:-1]          # remove some triggers if necessary, use [0:1, ...] to have consistent format
# triggers[0,2] = triggers[0,2][0:1, 1:-1]  

triggers = triggers_tot
clf.triggers_shape = triggers.shape
clf.triggers = triggers




# PEAK FINDER
# pf_height = params.classifier.peaks.height                # threshold of probability
# pf_width = params.classifier.peaks.width                     # consecutive bins above threshold
pf_distance = params.classifier.peaks.distance
if pf_distance == 0:
    pf_distance = params.recording.timing.t_response/fct         # distance between consecutive peaks

# PLOTTING PARAMETERS
# triggers_plot = triggers.shape[1]
triggers_plot = params.classifier.model_training.triggers_plot

if triggers_plot > params.classifier.n_classes:
    triggers_plot = params.classifier.n_classes



tp_tot = np.zeros((len(par_list), params.classifier.n_classes))       # tp, fp, tn
fp_tot = np.zeros((len(par_list), params.classifier.n_classes))       # tp, fp, tn
tn_tot = np.zeros((len(par_list), params.classifier.n_classes))       # tp, fp, tn
fn_tot = np.zeros((len(par_list), params.classifier.n_classes))       # tp, fp, tn

tpr_tot = np.zeros((len(par_list), params.classifier.n_classes))       # tp, fp, tn
fpr_tot = np.zeros((len(par_list), params.classifier.n_classes))       # tp, fp, tn
tnr_tot = np.zeros((len(par_list), params.classifier.n_classes))       # tp, fp, tn
fnr_tot = np.zeros((len(par_list), params.classifier.n_classes))       # tp, fp, tn
scikit_scores = []

# session_ids = list(range(0, data_tot.shape[0]))         # list with sessions
session_ids = list(range(0, 1))         # list with sessions for training


cnt = 0
for par_id, pars in enumerate(par_list):
    print(f'par_id: {par_id} out of {len(par_list)}')

    peaks_tot = []
    params.classifier.template = np.array(pars['template'])
    params.classifier.reg_fact = pars['reg_fact']
    pf_width = pars['pf_width'][0]
    pf_height = pars['pf_height']

    session_id = 0
    sess = 0
    # cur_data = data_tot[session_id:session_id + 1]
    cur_data = data_tot[-1:]          # get last session for testing

    tp = np.zeros((len(par_list), params.classifier.n_classes))
    fp = np.zeros((len(par_list), params.classifier.n_classes))
    tn = np.zeros((len(par_list), params.classifier.n_classes))
    fn = np.zeros((len(par_list), params.classifier.n_classes))

    model_fname1 = model_fname01 + '_' + dt.now().strftime('%H%M%S') + '.pkl'
    model_fname2 = model_fname02 + '_' + dt.now().strftime('%H%M%S') + '.pkl'

    # clf.get_trials_kiap(data_tot[session_ids, :], triggers, features_fname=features_fname)
    if data_tot.shape[0] > 1: 
        clf.get_trials_kiap(data_tot[:-1, :], triggers, features_fname=features_fname)                      # 1. GET TRAIN FEATURES
    else:
        clf.get_trials_kiap(data_tot[:, :], triggers, features_fname=features_fname)                      # 1. GET TRAIN FEATURES
    _, _, _, _ = clf.train_LDA(clf.train_trials, baseline=0, features_fname=features_fname,                     # 2. TRAIN MODEL
                        model_fname1=model_fname1, model_fname2=model_fname2, solver=solver, save_model=save_model)
                        # model_fname1=model_fname1, model_fname2=model_fname2, solver='lsqr', save_model=save_model)
    
    clf.get_test_trials_kiap(cur_data)                                                            # 3. GET TEST FEATURES
    clf.test_LDA_kiap(session_id=sess, compare_results=0)                                          # 4. TEST MODEL

    # plt.figure(1)
    # plt.clf()
    # # splot = plt.subplot(111)
    # # plot_ellipse(splot, clf.clf1.means_[0], clf.clf1.covariance_, 'r')
    # # plot_ellipse(splot, clf.clf1.means_[1], clf.clf1.covariance_, 'b')
    # # plt.xlim(-4,4)
    # # plt.ylim(-4,4)
    # plt.plot(clf.clf1.coef_.T)
    # plt.draw()
    # plt.show()
    # # print(clf.clf1.scores)
    # input()
    # xx

    scikit_scores.append(clf.clf1.scores)

    xx1 = np.arange(cur_data[0, 0].shape[0]) * fct
    xx2 = np.arange(clf.test_trials.shape[0]) * fct


    if 'explicit' in params.classifier.model_training.model:
        prob = clf.rel_prob     # explicit LDA
    elif 'scikit' in params.classifier.model_training.model:
        prob = clf.prob         # sklearn model

    if prob.shape[1] == 2:      # to be sure that beginning and end are also detected
        prob[0, 1] = 0
        prob[-1, 1] = 0

    
    pred0 = clf.clf1.predict(clf.test_trials)
    sig = np.zeros((pred0.size,triggers_plot))
    pred = np.zeros((pred0.size,triggers_plot))
    filtered = np.zeros((pred0.size,triggers_plot))

    # CALCULATE CLASSIFICATION PERFORMANCE
    for cl_id in range(triggers_plot):
        pred[pred0==cl_id,cl_id] = 1

        if 'prob' in params.classifier.peaks.sig:
            sig[:,cl_id] = prob[:,cl_id]
        elif 'pred' in params.classifier.peaks.sig:
            sig[:,cl_id] = pred[:,cl_id]

        if params.classifier.peaks.prefilter:
            filtered[:, cl_id] = signal.savgol_filter(sig[:, cl_id],51,3)
        else:
            filtered[:, cl_id] = sig[:, cl_id]

        # peaks, _ = signal.find_peaks(filtered[:,cl_id], height=pf_height, width=pf_width, distance=pf_distance)
        peaks = analytics1.detect_peaks(filtered[:,cl_id], mph=pf_height, mpd=pf_distance, edge='rising', width=pf_width)
        # print('peaks', peaks)

        # peaks_in = np.empty((0, 1), int)
        # for index in range(peaks.shape[0]):
        #     if np.any(filtered[peaks[index]:peaks[index] + pf_width, cl_id] < pf_height):
        #         peaks_in = np.append(peaks_in, index)
        # peaks = np.delete(peaks, peaks_in)
        # peaks += pf_width


        if params.classifier.trigger_pos == 'stop':

            peaks = peaks[peaks < xx2.size]

            temp = triggers[-1, cl_id][0,:]
            for idx in peaks:
                # if np.any(np.abs(xx2[idx] - triggers[0, cl_id] * fct) < params.recording.timing.t_response+max(pars['template'])/fct):
                diff1 = xx2[idx] - temp * fct
                if np.any(diff1[diff1<0] > -(params.recording.timing.t_response-min(pars['template'])*fct/2)):
                    temp = temp[1:]
                    tp[par_id, cl_id] += 1
                    # print(xx2[idx])
                else:
                    fp[par_id, cl_id] += 1

            fn[par_id, cl_id] = triggers[-1, cl_id].size - tp[par_id, cl_id]

            peaks_tot.append(peaks)

            tpr_tot[par_id, cl_id] = tp[par_id, cl_id] / triggers[-1, cl_id].size
            # FPR = FP/N = FP / (FP+TN)
            fpr_tot[par_id, cl_id] = fp[par_id, cl_id] / (fp[par_id, cl_id] + sum([triggers[-1, ii].size for ii in range(params.classifier.n_classes)]) - triggers[-1, cl_id].size)
            # fnr_tot[par_id, cl_id] = fn[par_id, cl_id] / (fn[par_id, cl_id] + tp[par_id, cl_id])
            fnr_tot[par_id, cl_id] = 1 - tpr_tot[par_id, cl_id]
            tnr_tot[par_id, cl_id] = 1 - fpr_tot[par_id, cl_id]

        else:

            peaks = peaks[peaks < xx2.size]

            temp = triggers[-1, cl_id][0,:]
            for idx in peaks:
                # if np.any(np.abs(xx2[idx] - triggers[0, cl_id] * fct) < params.recording.timing.t_response+max(pars['template'])/fct):
                diff1 = xx2[idx] - temp * fct
                if np.any(diff1[diff1>0] < params.recording.timing.t_response+max(pars['template'])*fct/2):
                    temp = temp[1:]
                    tp[par_id, cl_id] += 1
                    # print(xx2[idx])
                else:
                    fp[par_id, cl_id] += 1

            fn[par_id, cl_id] = triggers[-1, cl_id].size - tp[par_id, cl_id]

            peaks_tot.append(peaks)

            tpr_tot[par_id, cl_id] = tp[par_id, cl_id] / triggers[-1, cl_id].size
            # FPR = FP/N = FP / (FP+TN)
            fpr_tot[par_id, cl_id] = fp[par_id, cl_id] / (fp[par_id, cl_id] + sum([triggers[-1, ii].size for ii in range(params.classifier.n_classes)]) - triggers[-1, cl_id].size)
            # fnr_tot[par_id, cl_id] = fn[par_id, cl_id] / (fn[par_id, cl_id] + tp[par_id, cl_id])
            fnr_tot[par_id, cl_id] = 1 - tpr_tot[par_id, cl_id]
            tnr_tot[par_id, cl_id] = 1 - fpr_tot[par_id, cl_id]


    tp_tot[cnt, :] = tp[par_id]
    fp_tot[cnt, :] = fp[par_id]
    tn_tot[cnt, :] = tn[par_id]
    fn_tot[cnt, :] = fn[par_id]

    np.set_printoptions(precision=2, formatter={'int': lambda x: '%2d' % x, 'float': lambda x: '%.2f' % x}, )

    # for ii in range(cnt):
    #     # np.set_printoptions(precision=2)
    #     print(f'{par_list[ii]}\n')
    #     print('tp', tp_tot[ii,:], tpr_tot[ii,:])
    #     print('fp', fp_tot[ii,:], fpr_tot[ii,:])
    #     print('tn', tn_tot[ii,:], tnr_tot[ii,:])
    #     print('fn', tn_tot[ii,:], fnr_tot[ii,:])
    #     print()
    
    
        # assert(np.any(tpr_tot <=1))
        # np.testing.assert_array_almost_equal(tpr_tot[cnt], 1 - fnr_tot[cnt])
        # np.testing.assert_array_almost_equal(fpr_tot[cnt], 1 - tnr_tot[cnt])
        # print('Classification scores calculation is valid.\n')



    cnt += 1

acc = (tp_tot + tn_tot) / (tp_tot + tn_tot + fp_tot + fn_tot)
f1 = 2*tp_tot / (2*tp_tot + fp_tot + fn_tot)
acc_mean = np.mean(acc, axis=1)
f1_mean = np.mean(f1, axis=1)
scikit_scores_mean = np.mean(scikit_scores, axis=1)

print('SCORES\n______\n')
print(f'ACC: {acc}')
print(f'F1: {f1}')
print(f'scikit scores: {scikit_scores_mean}')

# a = np.argwhere(tp_tot[:, 0] == np.amax(tp_tot[:, 0]))
# b = np.argwhere(fp_tot[:, 0] == np.amin(fp_tot[:, 0]))
a = np.argwhere(acc_mean == np.amax(acc_mean))
b = np.argwhere(f1_mean == np.amax(f1_mean))
best_pars = [par_list[idx] for idx in np.intersect1d(a, b)]


print(f'\nBest pars:')
for ii in best_pars:
    print(ii)
print('\n')


idx = np.intersect1d(a, b)
print(f'\nBest pars idx: {idx}')
print('tp', tp_tot[idx,:], tpr_tot[idx,:])
print('fp', fp_tot[idx,:], fpr_tot[idx,:])
print('tn', tn_tot[idx,:], tnr_tot[idx,:])
print('fn', tn_tot[idx,:], fnr_tot[idx,:])
print(f'ACC: {acc[idx]}\nF1: {f1[idx]}\n')



np.set_printoptions(precision=8)


# PLOT RESULTS
# ------------

plt.figure(1)
plt.clf()
ax1 = plt.subplot(311)
ax2 = plt.subplot(312, sharex=ax1)
ax3 = plt.subplot(313, sharex=ax1)
plt.ylim(0,2)

plt.draw()


# ch_ids = range(0)
ch_ids = range(cur_data[0,0].shape[1])
# ch_ids = [93, 58, 70, 64, 78, 82, 43, 72, 86, 94]
# ch_ids = [0,1,2]

col = ['C0', 'C1', 'C2']

ymax1= np.max(cur_data[0,0][idx1:,ch_ids])
ymin1 = -1

plt.gca().set_prop_cycle(color=['#d62728', '#9467bd'])
# ax1.plot(xx1, cur_data[0, 0][:, 0:10],'-.', alpha=0.2)
# pca1 = PCA(n_components=2).fit_transform(cur_data[0,0])
# ax1.plot(xx1, cur_data[0, 0][:, ch_ids],'C0', alpha=0.4)
# ax1.plot(xx1, pca1[:,0],'C1', alpha=0.6)
# ax1.plot(xx1, pca1[:,1],'C3', alpha=0.6)

for cl_id in range(triggers_plot):
    markerline, stemlines, baseline = ax1.stem(triggers[-1, cl_id][0] * fct, triggers[-1, cl_id][0] * 0 + ymax1, '-', basefmt=' ', bottom=ymin1)
    plt.setp(stemlines, alpha=0.8, color=col[cl_id], lw=2)
    # plt.setp(markerline, marker=None)
    plt.setp(markerline, alpha=1, color=col[cl_id], lw=1)

    if peaks_tot[cl_id].size>0:
        markerline, stemlines, baseline = ax2.stem(xx2[peaks_tot[cl_id]], xx2[peaks_tot[cl_id]] * 0 + 2)
        plt.setp(stemlines, alpha=1, color=col[cl_id], lw=1)
        plt.setp(markerline, alpha=1, color=col[cl_id], lw=1)

    # filtered = signal.savgol_filter(pred[:, cl_id],51,3)
    ax2.plot(xx2, filtered[:,cl_id], '-', lw=1, color=col[cl_id])   
    # ax2.plot(xx2, pred[:,cl_id], '-', lw=1, color=col[cl_id])   


    ax3.plot(xx2, clf.rel_prob[:,cl_id], '-', lw=1, color=col[cl_id])


ax1.set_ylim(ymin1,ymax1*1.2)
# analytics1.plot_psth(clf)




plt.figure(2)
plt.clf()
plt.title('Scores')
plt.plot(acc_mean,'-o', label='ACC')
plt.plot(f1_mean,'-o', label='F1')
plt.plot(scikit_scores_mean, '-o', label = 'scikit')
plt.ylim(-0.1,1.1)
plt.legend()
plt.draw()
plt.show()


ans = input('Save last model and update config file y/n: ')


if ans =='y':
    comment = input('add short comment: ')
    clf.clf1.comment = comment
    clf.clf2.comment = comment
    
    out_conf = {'classifier': {
        'path_model1': f"{model_fname1}",
        'path_model2': f"{model_fname2}",
        'thr_prob': best_pars[0]['pf_height'][0],
        'thr_window': best_pars[0]['pf_width'][0],
        'reg_fact': float(best_pars[0]['reg_fact']),
        'template': best_pars[0]['template'].tolist()
    }}
    config_dump_stream = open(params.classifier.saved_model_conf_name, 'w')
    yaml.dump(out_conf, config_dump_stream, sort_keys=False, default_flow_style=None)
    

    print(f'Config file saved to: "{params.classifier.model_training.saved_model_conf_name}"')

    print(f'Saving model1 in {model_fname1} and \nmodel2 in {model_fname2}')
    joblib.dump(clf.clf1, model_fname1)       # scikit LDA
    joblib.dump(clf.clf2, model_fname2)       # explicit LDA