'''
description: parameter scan
author: Ioannis Vlachos
date:   13.12.18

Copyright (c) 2018  Ioannis Vlachos
All rights reserved.'''


import importlib
import warnings
from datetime import datetime as dt

# import copy
import matplotlib.pyplot as plt
# import pickle
import numpy as np
# import sys
import scipy.signal as signal
from scipy import io
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis as LDA
from sklearn.metrics import auc, classification_report, roc_curve
from sklearn.model_selection import (GridSearchCV, ParameterGrid,
                                     StratifiedShuffleSplit)

import analytics
import aux
import modules.classifier as clf
from aux import log
from helpers import data_management as dm

importlib.reload(aux)
importlib.reload(clf)
importlib.reload(dm)
importlib.reload(analytics)

aux.config_logging()

# param_grid = {'template': [[10,30,50]], 'reg_fact': np.arange(0.0, 1.01, .2)}
# param_grid = {'template': [np.arange(10,50).tolist()], 'reg_fact': np.arange(0.1, 0.11, .2)}
param_grid = {'template': [np.arange(1,10).tolist(),np.arange(1,20).tolist(), np.arange(1,50).tolist()]}
par_list = list(ParameterGrid(param_grid))


params = aux.load_config()
params.classifier.online = False
save_model = params.classifier.model_training

solver = params.classifier.model_training.solver

# features_fname = '/kiap/data/tom/train_trials.pkl'
features_fname = params.file_handling.datafile_path + '/train_trials.pkl'    # name of features array
model_fname01 = params.file_handling.datafile_path + '/model1'               # save sklearn LDA model
model_fname02 = params.file_handling.datafile_path + '/model2'               # save explicit LDA model


data_tot, tt, triggers = dm.get_raw(n_triggers=params.classifier.n_classes)
data_tot = analytics.exclude_channels(data_tot, params.classifier.channel_mask)         # exclude channels from analysis

session_ids = list(range(0, data_tot.shape[0]))         # list with sessions
clf = clf.Classifier(params)

# triggers[0,1] = triggers[0,1][0:1, 1:-1]          # remove some triggers if necessary, use [0:1, ...] to have consistent format
# triggers[0,2] = triggers[0,2][0:1, 1:-1]  

clf.triggers_shape = triggers.shape
clf.triggers = triggers


fct = params.daq.spike_rates.loop_interval / 1000.      # convert samples to sec

# PEAK FINDER
pf_threshold = params.classifier.peaks.height                # threshold of probability
pf_width = params.classifier.peaks.width                     # consecutive bins above threshold
pf_distance = params.recording.timing.t_response/fct         # distance between consecutive peaks
pf_distance = 10

# PLOTTING PARAMETERS
# triggers_plot = triggers.shape[1]
triggers_plot = 2




tp_tot = np.zeros((len(par_list), params.classifier.n_classes))       # tp, fp, tn
fp_tot = np.zeros((len(par_list), params.classifier.n_classes))       # tp, fp, tn
tn_tot = np.zeros((len(par_list), params.classifier.n_classes))       # tp, fp, tn

tpr_tot = np.zeros((len(par_list), params.classifier.n_classes))       # tp, fp, tn
fpr_tot = np.zeros((len(par_list), params.classifier.n_classes))       # tp, fp, tn
tnr_tot = np.zeros((len(par_list), params.classifier.n_classes))       # tp, fp, tn

peaks_tot = []

cnt = 0



scores = ['precision']
tuned_parameters = [{'shrinkage': list(np.arange(0.1,0.91,0.1)), 'solver': ['lsqr', 'eigen']}]
n_classes = 3

best_params = []
means = []
stds = []

for par_id, pars in enumerate(par_list):

    params.classifier.template = np.array(pars['template'])
    
    session_id = 0
    sess = 0
    cur_data = data_tot[session_id:session_id + 1]
    # cur_data = data_tot[-2:-1]          # get last session for testing

    tp = np.zeros((len(par_list), params.classifier.n_classes))
    fp = np.zeros((len(par_list), params.classifier.n_classes))
    tn = np.zeros((len(par_list), params.classifier.n_classes))

    model_fname1 = model_fname01 + '_' + dt.now().strftime('%H%M%S') + '.pkl'
    model_fname2 = model_fname02 + '_' + dt.now().strftime('%H%M%S') + '.pkl'

    clf.get_trials_kiap(data_tot[session_ids, :], triggers, features_fname=features_fname)                      # 1. GET TRAIN FEATURES
    data_train = clf.train_trials
    n_features = data_train[0, 0].shape[0]

    X_tot = np.empty((0, n_features))
    y_tot = np.empty((0, 1))


    for ii in range(n_classes):
        X_tot = np.vstack((X_tot, data_train[ii, 0].T))
        y_tot = np.vstack((y_tot, data_train[ii, 0][0][:, None] * 0 + ii))

    X_train = X_tot
    y_train = y_tot

    X_test = X_tot
    y_test = y_tot


    kf = StratifiedShuffleSplit(n_splits=10, test_size=0.2, random_state=None)

    for score in scores:
        print("# Tuning hyper-parameters for %s" % score)
        print()

        clf3 = GridSearchCV(LDA(), tuned_parameters, cv=kf, n_jobs=4,
                           scoring='%s_macro' % score, return_train_score=False)

        clf3.fit(X_train, y_train.ravel())

        # print(f'Best params: {clf3.best_params_}\n')
        best_params.append(clf3.best_params_)
        means.append((clf3.cv_results_['mean_test_score']).mean())
        stds.append((clf3.cv_results_['std_test_score']).mean())

        # for mean, std, model_params in zip(means, stds, clf3.cv_results_['params']):
            # print("%0.3f (+/-%0.03f) for %r"
                  # % (mean, std * 2, model_params))

        # y_pred = clf3.predict(X_test)
        # print(classification_report(y_test, y_pred))
        # print()

for ii, bpars in enumerate(best_params):
    print(f'{ii}, {bpars}, {means[ii]:.2f}, {stds[ii]:.2f}')
