'''
description: script compares python decoder vs matlab decoder for the eight sessions provided by Tom
author: Ioannis Vlachos
date:   14.11.18

Copyright (c) 2018  Ioannis Vlachos
All rights reserved.'''


import copy
import importlib
import pickle
import sys

import matplotlib.pyplot as plt
import numpy as np
from scipy import io

import aux
import modules.classifier as clf
from aux import log

importlib.reload(aux)
importlib.reload(clf)

get_features = int(sys.argv[1])
compute_model = int(sys.argv[2])

params = aux.load_config()


# features_fname = '/kiap/data/tom/train_trials.pkl'
features_fname = params.file_handling.datafile_path + 'train_trials.pkl'
model_fname1   = params.file_handling.datafile_path + 'model1.pkl'      # save sklearn LDA model
model_fname2   = params.file_handling.datafile_path + 'model2.pkl'      # save explicit LDA model

data_tot = io.loadmat('/kiap/data/tom/model/trainData_rates2.mat')['trainData']         # SPIKERATES: (n_train_sess, 1) with (n_samples, n_ch) per element
triggers = io.loadmat('/kiap/data/tom/model/trainData_triggers.mat')['trainTrig']       # (n_train_sess, n_ triggers)
train_trials_comparison = io.loadmat('/kiap/data/tom/model/trainTrials.mat')['trainTrials']    # FEATURES:   (n_classes, 1)  with (n_feat, n_trials) per class


session_ids = list(range(0, 8))         # list with sessions
clf1 = clf.Classifier(params)
clf1.triggers_shape = triggers.shape


if get_features:
    # log.info('Get feature array from new spike-rate data')
    clf1.get_trials(data_tot[session_ids, :], triggers, features_fname=features_fname, train_trials_comparison=train_trials_comparison)
else:
    log.info('Old feature array will be used')

if compute_model:
    log.info('Computing classifier model...')
    _, _, _, _ = clf1.train_LDA(clf1.train_trials, features_fname=features_fname,
                model_fname1=model_fname1, model_fname2=model_fname2, solver='svd', save_model=1)
else:
    log.info('Stored classifier model will be used.')


plt.clf()


session_id = 7

for sess in range(session_id, session_id + 1):
    print(sess)
    cur_data = data_tot[session_id:session_id + 1]

    test_trials = clf1.get_test_trials(cur_data)

    clf1.test_LDA(test_trials, session_id=sess, compare_results=1)

    for jj in range(600, 4000, 20):
        clf1.get_class(cur_data[0, 0][jj - 600:jj], jj)
