'''
description: implement actual decoder based on rLDA
author: Ioannis Vlachos
date:   20.08.18

Copyright (c) 2018 Ioannis Vlachos.
All rights reserved.'''

import pickle

import numpy as np
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis as LDA

classes = np.load('classes.npz')
class1 = classes['class1']
class2 = classes['class2']

Xtot = np.vstack((class1, class2))
ytot = np.vstack((np.zeros((pars.trials.n_bl, 1)), np.ones((pars.trials.n_cues, 1))))

idx = np.random.permutation(range(Xtot.shape[0]))

Xtot = Xtot[idx]
ytot = ytot[idx]


ntrain = int(0.7 * Xtot.shape[0])
Xtrain = Xtot[:ntrain, :]
ytrain = ytot[:ntrain]

Xtest = Xtot[ntrain:, :]
ytest = ytot[ntrain:]

clf = LDA(solver='svd', shrinkage=None).fit(Xtrain, ytrain)
score1 = clf.score(Xtrain, ytrain)
score2 = clf.score(Xtest, ytest)
print("\nTraining and test scores: {:.2f}, {:.2f}".format(score1, score2))


with open('classifier.pkl', 'wb') as fid:
    pickle.dump(clf, fid)
