
'''
description: implementation of an LDA. confirmed on real data  - uses shrinkage algorithm
author: Ioannis Vlachos
date:   23.04.18

Copyright (c) 2018 Ioannis Vlachos.
All rights reserved.
'''

import random
import sys

import numpy as np
import scipy
from scipy import io
from sklearn.covariance import OAS, LedoitWolf
from sklearn.discriminant_analysis import _cov as lda_cov


def LDA(X, y, train_id, n_features):

    Xtrain = X[:train_id,:n_features]
    ytrain = y[:train_id]

    Xtest = X[train_id:,:n_features]
    ytest = y[train_id:]


    _, y_t = np.unique(ytrain, return_inverse=True)
    priors = np.bincount(y_t) / float(len(ytrain))


    n_cl = 2
    # compute mean vectors
    means = []
    for cl in range(n_cl):
        means.append(np.mean(Xtrain[ytrain==cl], axis=0))
        # print('Mean Vector class %s: %s\n' %(cl, means[cl]))



    # compute within-class scatter matrices - using np.cov
    cov_cl = []
    for cl_id in range(n_cl):       
        # cov_cl.append(np.cov(Xtrain[ytrain==cl_id].T,bias=1)) 
        cov_cl.append(lda_cov(Xtrain[ytrain==cl_id],shrinkage='auto')) 
        # oa = OAS()                                                        % takes more time but does not improve much
        # cov_cl.append(oa.covariance_) 

    Sw = np.average(cov_cl,axis=0, weights = priors)


    # compute between-class scatter matrices - using np.cov
    # St = np.cov(Xtrain.T,bias=1)
    St = lda_cov(Xtrain, shrinkage='auto')
    Sb = St - Sw



    try:
        evals, evecs = scipy.linalg.eigh(Sb,Sw)             # see sklearn
        evecs = evecs[:, np.argsort(evals)[::-1]]
        evecs /= np.linalg.norm(evecs, axis=0)
    except:
        print('eigenvector computation error')
        return(np.nan,np.nan)

    W = evecs


    # compute decision function


    coef = np.dot(means, W).dot(W.T)
    intercept = -0.5 * np.diag(np.dot(means, coef.T)) + np.log(priors)


    scores1 = np.dot(Xtrain, coef.T) + intercept
    scores2 = np.dot(Xtest, coef.T) + intercept

    yp1 = scores1.argmax(axis=1)
    yp2 = scores2.argmax(axis=1)


    perf1 = (yp1.size - np.count_nonzero(ytrain-yp1))/np.float(yp1.size)
    perf2 = (yp2.size - np.count_nonzero(ytest-yp2))/np.float(yp2.size)

    return perf1,perf2
