import importlib
import sys
import warnings

import matplotlib.pyplot as plt
import numpy as np
from scipy import fftpack, signal

import aux
import colorednoise as cn
import decoder as dec
import methods

importlib.reload(dec)

warnings.simplefilter(action='ignore', category=FutureWarning)


fc1 = 5        # signal 1  freq
fc2 = 25        # signal 2  freq

# fs = 4. * (2 * np.pi * max(fc1, fc2))   # sampling rate

n_bins = 4                              # number of bins to group freq bands
fct = 30                                # downsampling factor
A = .1         # white noise amplitude


nfft_fct = 1


N = 50
n_trials = N
n_trials_bl = N
tlen = 3        # sec

n_exp = 1     # colored noise exponent


def generate_signal(response=0, N=1):

    '''response: 1: generate signal, 0: only noise
    N: number of trials'''


    fs = params.sampling.fs                   # sampling rate
    ts = 1. / fs                            # sampling period

    tt0 = np.arange(0, tlen, ts)                           # time points - (nsamples,)
    tt1 = np.repeat(tt0[None, :], n_ch, axis=0)         # time points - (n_ch, nsamples)
    tidx1 = tt0.size // 2

    arr00 = np.sin(2 * np.pi * fc1 * tt1) * 1. / fc1**n_exp
    arr01 = np.sin(2 * np.pi * fc2 * tt1) * 1. / fc2**n_exp
    arr00[:, tidx1:] = 0
    arr01[:, :tidx1] = 0
    arr0 = arr00 + arr01

    arr = np.zeros((n_ch, tt0.size, N))

    for ii in range(N):
        arr[:, :, ii] = arr0 * response + generate_colored_noise(arr0.shape)*0.2
    arr2, tt2 = signal.resample(arr, arr.shape[1] // 30, t=tt0, axis=1)
    fs = fs / fct

    if response:
        plt.plot(arr2[0, :, :])
    return arr2, tt2, fs, arr



def generate_colored_noise(shape):
    beta = 1             # the exponent
    n_samples = np.prod(shape)
    noise = cn.powerlaw_psd_gaussian(beta, n_samples)
    noise = noise.reshape(shape)
    return noise

def get_stft(arr):
    ff, tt, sp_amp = signal.stft(arr, fs=fs, window='hamming', nperseg=256, noverlap=203, axis=1)
    sp_amp = np.moveaxis(sp_amp, 2, 3)  # move trials to end
    return ff, tt, np.abs(sp_amp)


def get_fft(arr):
    sp_amp = fftpack.fft(arr, n=arr.shape[1] * nfft_fct, axis=1)
    sp_amp = sp_amp / sp_amp.shape[1] * 2
    ff = fftpack.fftfreq(sp_amp.shape[1], d=1. / fs)
    sp_amp = np.abs(sp_amp)
    fidx = ff > 0
    return ff[fidx], sp_amp[:, fidx, :]



def get_bands(arr, arr2, f, params):

    arr = nspat          # ch x fbins x tbins x trials
    arr2 = nspat_bl     # ch x fbins x tbins x trials       todo: check if it makes sense to normalzie basline
    f = ff_spat
    bb = f.size     # band_tot: f1, f2, ch, tbins, trials

    fmax = 200
    fidx = f < fmax
    fnum = f[fidx].size


    # if calc_bands:
        
    band_tot = np.zeros((fnum, fnum, params.general.n_channels, arr.shape[2], arr.shape[3]))
    band_tot_bl = np.zeros((fnum, fnum, params.general.n_channels, arr.shape[2], arr.shape[3]))
    for ii in range(fnum):
        print(ii)
        for jj in range(fnum):
            if jj > ii:
                idx = (f >= f[ii]) & (f < f[jj])
                band_tot[ii, jj, :, :] = np.sum(arr[:, idx, :, :], axis=1) / (f[jj] - f[ii])
                band_tot_bl[ii, jj, :, :] = np.sum(arr2[:, idx, :, :], axis=1) / (f[jj] - f[ii])


    band_tot_bl1 = np.mean(band_tot_bl, axis=3)     # average across time bins
    band_tot_bl2 = np.repeat(band_tot_bl1[:, :, :, None, :], band_tot_bl.shape[3], axis=3)    # repeat same value across time


    mu_cue = np.mean(band_tot, axis=4)     # average accross trials
    mu_bl = np.mean(band_tot_bl, axis=4)     # average accross trials
    std_cue = np.std(band_tot, axis=4)     # average accross trials
    std_bl = np.std(band_tot_bl, axis=4)     # average accross trials

    snr = (mu_cue - mu_bl) / (std_cue + std_bl)
    # snr = mu_cue
    snr2 = np.nanmax(snr, axis=3)
    snr3 = np.mean(snr2, axis=2)

    return snr3, f[fidx]



# ------------------------------------------------------------------------------------------------
# MAIN PART
# ------------

params = aux.load_config()
n_ch = params.general.n_channels

generate_data = int(sys.argv[1])        # 0: don't generate data to save time

if generate_data:
    bl, _, fs, _ = generate_signal(response=0, N=n_trials_bl)           # generate baseline data
    bl, taps = methods.high_pass_filter(bl, fs, params.highpass.fc)
    # methods.filter_response(taps, fs, params.lowpass.fc, 800)
    methods.filter_response(taps, fs, params.highpass.fc, 80)
    _, _, spat_bl = get_stft(bl)
    ff_spa_bl, spa_bl = get_fft(bl)

    arr, tt, fs, arr0 = generate_signal(response=1, N=n_trials)         # generate historical response data
    arr_hp, taps = methods.high_pass_filter(arr, fs, params.highpass.fc)
    ff_spat, tt_spat, spat = get_stft(arr_hp)
    ff_spa, spa = get_fft(arr_hp)


# normalize
spat_bl_norm0 = np.mean(spat_bl, axis=2)     # average across time bins
spat_bl_norm1 = np.repeat(spat_bl_norm0[:, :, None, :], spat.shape[2], axis=2)    # repeat same value across time
nspat = spat / spat_bl_norm1
nspat_bl = spat_bl / spat_bl_norm1


snr, ff = get_bands(nspat, nspat_bl, ff_spat, params)           # sp_tft should be: ch x fbins x tbins

# fidx = ff_spat < 100


ch_id = 0
trial_id = 0


np.save('/kiap/data/simulated/arr.npy', arr)
# ------------------------------------------------------------------------------------------------
# PLOT RESULTS
# ------------
plt.figure(1)
plt.clf()
plt.subplot(211)
plt.plot(tt.flatten(), arr_hp[0, :, :])
plt.subplot(212)
plt.plot(ff_spa, spa[ch_id, :, :], '--', lw=1, alpha=0.5)
# plt.plot(ff_spa, np.mean(spa, axis=2)[ch_id, :], label='mean')
plt.plot(ff_spa, np.mean(spa, axis=2)[ch_id, :] / np.mean(spa_bl, axis=2)[ch_id, :], label='mean2')
plt.legend(loc=0)
plt.xlim(0, 300)
plt.draw()
plt.show()


plt.figure(2)
plt.clf()
plt.imshow(np.flipud(snr), extent=[0, ff[-1], 0, ff[-1]], interpolation='bilinear')
plt.xlim(0, 100)
plt.ylim(0, 100)
plt.colorbar()

plt.tight_layout()
plt.draw()
plt.show()
