import importlib
import sys
import time
import warnings

import matplotlib.pyplot as plt
import numpy as np
from scipy import fftpack, signal

import aux
import colorednoise as cn
import decoder as dec
import methods

importlib.reload(methods)


nfft_fct = 1
ch_id = 0

ii = 0
twin = 500  # msec, time window for analysis
tstep = 50  # msec, update interval
n_taps = 12
plt_win = 2000


def get_fft(arr):
    sp_amp = fftpack.fft(arr, n=arr.shape[1] * nfft_fct, axis=1)
    sp_amp = sp_amp / sp_amp.shape[1] * 2
    ff = fftpack.fftfreq(sp_amp.shape[1], d=1. / fs)
    sp_amp = np.abs(sp_amp)
    fidx = ff > 0
    return ff[fidx], sp_amp[:, fidx]


params = aux.load_config()
n_ch = params.general.n_channels
fs = params.sampling.fs / params.sampling.smpl_fct
arr = np.load('/kiap/data/simulated/arr.npy')

# load baseline data
bl = np.load('/kiap/data/simulated/baseline.npy')
bl, taps = methods.high_pass_filter(bl, fs, params.highpass.fc)
# methods.filter_response(taps, fs, params.lowpass.fc, 800)
ff_fft_bl, sp_fft_bl = get_fft(bl)




plt.figure(6)
plt.clf()
ax1 = plt.subplot(311)
plt.xlim(0, plt_win)

ax2 = plt.subplot(312)
plt.xlim(0, plt_win)
# plt.ylim(0, 1)

plt.subplot(313)
plt.xlim(0, 300)
# plt.ylim(0, 0.2)
plt.ion()
plt.draw()
plt.show()

trial_id = -1

try:
    # Put matplotlib.pyplot in interactive mode so that the plots are shown in a background thread.
    plt.ion()
    while(True):
        input('Press enter to start next trial')
        trial_id += 1
        ii = 0
        ax1.cla()
        ax2.cla()
        plt.axes(ax1)
        plt.title(f'trial: {trial_id}')
        while ii < arr.shape[1]  - twin:
            stream = arr[:, ii:ii + twin, trial_id]
            # print(ii, ii + twin)
            ii = ii + tstep
            arr_hp, taps = methods.high_pass_filter(stream, fs, params.highpass.fc)
            # ff_stft, tt_stft, sp_stft = get_stft(arr_hp)
            ff_fft, sp_fft = get_fft(arr_hp)

            sp_fft = sp_fft / np.mean(sp_fft_bl[:, :sp_fft.shape[1]], axis=2)

            tt = np.arange(ii, ii+twin)

            fc1 = 0, 10     # frequency component 1, low, high
            fc2 = 25, 35    # frequency component 2, low, high
            fidx1 = (ff_fft > fc1[0]) & (ff_fft < fc1[1])
            fidx2 = (ff_fft > fc2[0]) & (ff_fft < fc2[1])
            # plt.clf()
            plt.subplot(311)
            plt.plot(tt[n_taps:], arr_hp[0, n_taps:], 'b')
            plt.xlim(max(0, ii - plt_win // 2), max(ii+plt_win // 2, plt_win))

            plt.subplot(312)
            # fidx = ff_fft
            # plt.plot(ii+twin, sp_fft[ch_id, ff_fft[fidx1]][None, :], 'b.')
            # plt.plot(ii+twin, sp_fft[ch_id, ff_fft[fidx2]][None, :], 'm.')
            plt.plot(ii+twin, np.sum(sp_fft[ch_id, fidx1], axis=0) / np.diff(fc1), 'b.')
            plt.plot(ii+twin, np.sum(sp_fft[ch_id, fidx2], axis=0) / np.diff(fc2), 'r.')
            plt.xlim(max(0, ii-plt_win//2), max(ii+plt_win//2, plt_win))

            plt.subplot(313)
            # plt.cla()
            lines = plt.plot(ff_fft, sp_fft[ch_id, :], 'b', lw=2, alpha=1)
            # plt.legend(loc=0)
            # time.sleep(.5)
            # plt.draw()
            plt.pause(0.001)
            lines.pop(0).remove()
            # plt.show()

except KeyboardInterrupt:
    # sys.exit(0)
    pass
