'''
description: test connection to NSP via cereconn
author: Ioannis Vlachos
date:   15.11.18

Copyright (c) 2018 Ioannis Vlachos.
All rights reserved.'''


import time

import cere_conn as cc

ck = cc.CereConn()

ck.send_open()

t = time.time()
while ck.get_state() != cc.ccS_Idle:
    time.sleep(0.005)
print("It took {:5.3f}s to open CereConn\n".format(time.time() - t))

ck.send_record()
t = time.time()
while ck.get_state() != cc.ccS_Recording:
    time.sleep(0.02)
print("It took {:5.3f}s to start CereConn recording\n".format(time.time() - t))
cb_time = ck.get_cb_time()
print(cb_time)

t = time.time()
while (time.time()-t < 1):
    print(ck.get_cb_time()['ts']/1000.)
    time.sleep(0.1)


n_comments = 1

start_time = time.time()

round_trip_times = []
round_trip_times_abs = []

ck.get_comment_data()
for i in range(n_comments):
    cb_time = ck.get_cb_time()
    ck.send_comment("Test {} at tick {}".format(i+1, cb_time), cb_time)
    cd = ck.get_comment_data()
    while cb_time not in [c['data'] for c in cd['comments'] ]:
        time.sleep(0.001)
        cd = ck.get_comment_data()
    com = [com for com in cd['comments'] if com['data'] == cb_time]
    round_trip_times.append(com[0]['ts'] - com[0]['data'])
    round_trip_times_abs.append(cd['ts'] - com[0]['data'])

    
print("Average round trip times for {} comments: {} ticks, or {:5.1f}ms (min {:5.1f}, max {:5.1f})".format(n_comments, sum(round_trip_times)/len(round_trip_times), sum(round_trip_times)/len(round_trip_times) / 30.0, min(round_trip_times)/30.0, max(round_trip_times)/30.0 ))
print("Average round trip times for {} comments until processed here: {} ticks, or {:5.1f}ms (min {:5.1f}, max {:5.1f})".format(n_comments, sum(round_trip_times_abs)/len(round_trip_times_abs), sum(round_trip_times_abs)/len(round_trip_times_abs) / 30.0, min(round_trip_times_abs)/30.0, max(round_trip_times_abs)/30.0 ))


ck.send_close()
t = time.time()
while ck.get_state() != cc.ccS_Closed:
    time.sleep(0.05)

print("It took {:5.3f}s to close connection to NSP\n".format(time.time() - t))
