'''
description: script to read and write header information to file
author: Ioannis Vlachos
date:   02.11.18

Copyright (c) 2018 Ioannis Vlachos.
All rights reserved.'''


import ctypes
import multiprocessing as mp
from datetime import datetime as dt

import numpy as np
from numpy import datetime64 as dt64

shared_array_base = mp.Array(ctypes.c_float, 300 * 100 * 96)
data_buffer = np.ctypeslib.as_array(shared_array_base.get_obj())
db = data_buffer.reshape(300 * 100, 96)

np.random.seed(0)

# t_now1 = np.zeros((1), dtype='datetime64[us]')
# t_now2 = np.zeros((1), dtype='datetime64[us]')


with open('buffer.bin', 'wb') as fh:      # data file
    
    for ii in range(10):
        t_now1 = np.array(dt.now(), dtype='datetime64[us]')
        t_now2 = np.array(dt.now(), dtype='datetime64[us]')
        fh.write(t_now1.tobytes())       # one from NSP
        fh.write(t_now2.tobytes())       # one from PC

        db[:] = np.random.randint(0, 100, (30000, 96))
        db_bytes = db.tobytes()

        n_bytes = np.int32(len(db_bytes))
        n_samples = np.int32(db.shape[0])
        n_ch = np.int32(db.shape[1])
        fh.write(n_bytes)
        fh.write(n_samples)
        fh.write(n_ch)
        fh.write(db_bytes)
        print(ii, t_now1, t_now2, n_bytes, n_samples, n_ch, db[0, :10])

print('wrote to file done\n')


np.random.seed(0)
ii = 0
with open('buffer.bin', 'rb') as fh:
    while True:
        
        db[:] = np.random.randint(0, 100, (30000, 96))

        t_now1 = np.frombuffer(fh.read(8), dtype='datetime64[us]')

        if t_now1.size == 0:
            break

        t_now2 = np.frombuffer(fh.read(8), dtype='datetime64[us]')
        n_bytes = int.from_bytes(fh.read(4), byteorder='little')
        n_samples = int.from_bytes(fh.read(4), byteorder='little')
        n_ch = int.from_bytes(fh.read(4), byteorder='little')
        d = fh.read(n_bytes)
        d2 = np.frombuffer(d, dtype=np.float32)
        d3 = d2.reshape(d2.size // n_ch, n_ch)
        
        print(ii, t_now1, t_now2, n_bytes, n_samples, n_ch, d3[0, :10])

        assert(np.array_equiv(db, d3))

        ii +=1


print('read from file done')
