"""
Test Module for CBPY

Total of 4 tests that ensure that the correct library is being used to
negotiate with the NSP, and that channel configuration is working properly.

@Author:    Ali Zaidi
@Date:      18.09.2018
"""


import time
import unittest

import cerebus.cbpy as cb


class TestNSP(unittest.TestCase):

    @classmethod
    def setUp(self):
        self.channels = list(range(1, 97))

    def test_1_versionComptability(self):
        print("Testing version compatibility between NSP and CBSDK library")
        ver = cb.version()
        self.assertEqual(ver[1].get('major'), ver[1].get('nsp_major'))
        self.assertEqual(ver[1].get('minor'), ver[1].get('nsp_minor'))

    # def test_2_channelConfigWithDifferentSamplingRates(self):
    #     print("Configuring 5 channels into different sampling groups")
    #     for channel in self.channels:
    #         res = cb.set_channel_config(channel, chaninfo={'smpgroup': channel})
    #         # time.sleep(0.1)

    #     cb.trial_config(reset=True, buffer_parameter={'continuous_length': 30001})
    #     time.sleep(0.2)
    #     dat = cb.trial_continuous(reset=True)
    #     self.assertEqual(dat[1][0][2], 500)
    #     self.assertEqual(dat[1][1][2], 1000)
    #     self.assertEqual(dat[1][2][2], 2000)
    #     self.assertEqual(dat[1][3][2], 10000)
    #     self.assertEqual(dat[1][4][2], 30000)

    def test_3_channelConfigWith30kBufferSize(self):
        print("Configuring 96 channels at 30k and buffer size to 30000 samples")

        for channel in self.channels:
            res = cb.set_channel_config(channel, chaninfo={'smpgroup': 5})
            time.sleep(0.02)

        res = cb.trial_config(reset=True, buffer_parameter={'continuous_length': 30001})
        time.sleep(1.1)
        print("Ensuring channels return data with 30k samples")

        dat = cb.trial_continuous(reset=True)
        for channel in self.channels:
            self.assertEqual(len(dat[1][channel-1][1]), 30000)

    # def test_4_nChansReturned(self):
    #     print("Ensuring only channels that have been configured return data")
    #     cb.trial_config(reset=True)
    #     time.sleep(1)
    #     dat = cb.trial_continuous(reset=True)

    #     self.assertEqual(len(dat[1]), 5)

    # def test_5_nspResponseTime(self):
    #     import timeit
    #     s = "cb.trial_data()"
    #     time = timeit.timeit(stmt=s, number=10000, setup="from __main__ import cb")
    #     self.assertLessEqual(time, 50.)

    # def test_6_wrongCommandToNsp(self):
    #     with self.assertRaises(AttributeError):
    #         res = cb.trial_cont()


    # def test_7_wrongParamtoNsp(self):
    #     with self.assertRaises(TypeError):
    #         res = cb.trial_config(reset=True, buffer_parameters={'continuous': 10000})

    # @classmethod
    # def tearDown(self):
    #     pass


if __name__ == '__main__':
    import sys

    print("\n\nSTARTING TEST SUITE \n\n")

    res = cb.open()
    print("NSP instance number = {}".format(res[0]))
    if res[0] <= 0:
        sys.exit("Could not connect to NSP! Please make sure NSP is online")

    unittest.main(verbosity=2)

    cb.close()
