'''
description: testing prediction and speller matrix by writing freely on the console
author: Alessandro Tonin
date:   06.02.2020

Copyright (c) 2020 Alessandro Tonin.
All rights reserved.'''

import importlib
import logging
import sys
import time
from random import randint

import tkinter as tk
import munch
# import pyaudio
from colorlog import ColoredFormatter

import aux
from aux import log
from paradigms import colorSpeller, exploration, question, trainingColorSpeller, sp_matrix
import multiprocessing as mp
from multiprocessing import Array, Pipe

import readchar

importlib.reload(colorSpeller)
importlib.reload(aux)

def speller_matrix(child_conn):
    root = tk.Tk()
    sp_matrix.sp_matrix(root, child_conn)
    root.mainloop()

parent_conn, child_conn = Pipe()

matrix = mp.Process(name='matrix', target=speller_matrix, args=(child_conn,))
matrix.daemon = True    
matrix.start()

params = aux.load_config()

sp = colorSpeller.ColorSpeller(pyttsx_rate = params.speller.pyttsx_rate, params=params)

finish = False
parent_conn.send(['', sp.current_string, sp.get_vocabulary()])

while not finish:

	# space: ' '
	# full stop: '.'
	# backspace: '\x7f'
	# delete word: '`'
	# quit (enter): '\r'
	c = readchar.readchar()
	if c == '\x7f':
		selection = '<'
		sp.set_current_string(sp.current_string[:-1])
	elif c == '`':
		selection = '<<<'
		all_words = sp.current_string.split()[:-1]
		new_str = ' '.join(all_words)
		if new_str:
			new_str += ' '
		sp.set_current_string(new_str)
	elif c == '\r':
		selection = 'end'
		break
	elif c == '.':
		selection = '.'
		sp.set_current_string(sp.current_string + '. ')
	elif c == '\t':
		if sp.get_vocabulary(1)[0].startswith('*'):
			selection = sp.get_vocabulary(1)[0]
			all_words = sp.current_string.split()[:-1]
			all_words.append(selection[1:-1])
			new_str = ' '.join(all_words)
			sp.set_current_string(new_str+' ')
		else:
			continue
	else:
		selection = c.upper()
		sp.set_current_string(sp.current_string+selection)

	sp.external_updates()

	log.warning(f'string: {sp.current_string}')
	log.warning(f'string: {sp.get_vocabulary()}')

	parent_conn.send([selection, sp.current_string, sp.get_vocabulary()])

matrix.terminate()